@@ -38,10 +38,10 @@ use std::{borrow::Cow, sync::Arc};
 use indexmap::IndexMap;
 use melib::{
     backends::{AccountHash, BackendEvent, MailboxHash},
-    error::Error,
-    EnvelopeHash, RefreshEvent, RefreshEventKind, ThreadHash,
+    error::{Error, Result, ResultIntoError},
+    log, EnvelopeHash, RefreshEvent, RefreshEventKind, ThreadHash,
 };
-use nix::unistd::Pid;
+use nix::{errno::Errno, unistd::Pid};
 
 use super::{
     command::Action,
@@ -129,9 +129,7 @@ impl From<UIEvent> for ThreadEvent {
 }
 
 #[derive(Debug)]
-pub enum ForkType {
-    /// Already finished fork, we only want to restore input/output
-    Finished,
+pub enum ForkedProcess {
     /// Embedded pty
     Embedded {
         id: Cow<'static, str>,
@@ -145,6 +143,161 @@ pub enum ForkType {
     },
 }
 
+impl ForkedProcess {
+    /// Return the process's string identifier.
+    #[inline]
+    pub fn id(&self) -> &str {
+        let (Self::Embedded { id, .. } | Self::Generic { id, .. }) = self;
+        id.as_ref()
+    }
+
+    /// Return the process's ID.
+    pub fn pid(&self) -> Pid {
+        match self {
+            Self::Embedded { pid, .. } => *pid,
+            Self::Generic { child, .. } => nix::unistd::Pid::from_raw(child.id() as libc::pid_t),
+        }
+    }
+
+    /// Attempt to kill process.
+    pub fn kill(&mut self) -> Result<()> {
+        match self {
+            Self::Generic {
+                ref id,
+                ref command,
+                ref mut child,
+            } => {
+                let w = child.kill();
+                match w {
+                    Ok(()) => {
+                        log::debug!(
+                            "child process {} {} ({}) was killed successfully",
+                            id,
+                            command
+                                .as_deref()
+                                .map(Cow::Borrowed)
+                                .unwrap_or(Cow::Borrowed("<command unknown>")),
+                            child.id(),
+                        );
+                        Ok(())
+                    }
+                    err @ Err(_) => err.chain_err_details(|| {
+                        format!(
+                            "Failed to kill child process {} {} ({})",
+                            id,
+                            command
+                                .as_deref()
+                                .map(Cow::Borrowed)
+                                .unwrap_or(Cow::Borrowed("<command unknown>")),
+                            child.id(),
+                        )
+                    }),
+                }
+            }
+            Self::Embedded {
+                ref id,
+                ref command,
+                ref pid,
+            } => {
+                use nix::sys::signal::{kill, Signal};
+                match kill(*pid, Some(Signal::SIGTERM)) {
+                    Ok(()) | Err(Errno::ESRCH) => Ok(()),
+                    err @ Err(_) => err.chain_err_details(|| {
+                        format!(
+                            "Failed to SIGTERM embedded child process {} {} ({})",
+                            id,
+                            command
+                                .as_deref()
+                                .map(Cow::Borrowed)
+                                .unwrap_or(Cow::Borrowed("<command unknown>")),
+                            pid,
+                        )
+                    }),
+                }
+            }
+        }
+    }
+
+    /// Try wait on child process without blocking.
+    ///
+    /// Return value is:
+    ///
+    /// - `Ok(true)` if the child has exited.
+    /// - `Ok(false)` if the child is still running.
+    /// - an error value, otherwise.
+    pub fn try_wait(&mut self) -> Result<bool> {
+        match self {
+            Self::Generic {
+                ref id,
+                ref command,
+                ref mut child,
+            } => match child.try_wait() {
+                Ok(Some(exit_status)) => {
+                    log::debug!(
+                        "child process {} {} ({}) exit_status = {:?}",
+                        id,
+                        command
+                            .as_deref()
+                            .map(Cow::Borrowed)
+                            .unwrap_or(Cow::Borrowed("<command unknown>")),
+                        child.id(),
+                        exit_status
+                    );
+                    Ok(true)
+                }
+                Ok(None) => Ok(false),
+                Err(err) => Err(err).chain_err_details(|| {
+                    format!(
+                        "Failed to wait on child process {} {} ({})",
+                        id,
+                        command
+                            .as_deref()
+                            .map(Cow::Borrowed)
+                            .unwrap_or(Cow::Borrowed("<command unknown>")),
+                        child.id(),
+                    )
+                }),
+            },
+            Self::Embedded {
+                ref id,
+                ref command,
+                ref pid,
+            } => {
+                use nix::sys::wait::{waitpid, WaitPidFlag, WaitStatus};
+                match waitpid(*pid, Some(WaitPidFlag::WNOHANG)) {
+                    Err(Errno::ECHILD) => Ok(true),
+                    Err(Errno::EAGAIN) => Ok(false),
+                    Err(err) => Err(err).chain_err_details(|| {
+                        format!(
+                            "Failed to wait on child process {} {} ({})",
+                            id,
+                            command
+                                .as_deref()
+                                .map(Cow::Borrowed)
+                                .unwrap_or(Cow::Borrowed("<command unknown>")),
+                            pid,
+                        )
+                    }),
+                    Ok(ws @ WaitStatus::Exited(_, _)) | Ok(ws @ WaitStatus::Signaled(_, _, _)) => {
+                        log::debug!(
+                            "embedded child process {} {} ({}) wait status: {:?}",
+                            id,
+                            command
+                                .as_deref()
+                                .map(Cow::Borrowed)
+                                .unwrap_or(Cow::Borrowed("<command unknown>")),
+                            pid,
+                            ws,
+                        );
+                        Ok(true)
+                    }
+                    Ok(_) => Ok(false),
+                }
+            }
+        }
+    }
+}
+
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 pub enum NotificationType {
     Info,
@@ -174,7 +327,10 @@ pub enum UIEvent {
     InsertInput(Key),
     EmbeddedInput((Key, Vec<u8>)),
     Resize,
-    Fork(ForkType),
+    Fork(ForkedProcess),
+    /// Restore input/output, useful after forking to something that captures
+    /// stdin/stdout.
+    RestoreStandardIO,
     ChangeMailbox(usize),
     ChangeMode(UIMode),
     Command(String),
