@@ -38,6 +38,7 @@ pub mod prelude {
         borrow::Cow,
         cell::RefCell,
         collections::{BTreeSet, HashMap},
+        convert::TryFrom,
         ops::Deref,
         pin::Pin,
         sync::Arc,
@@ -47,15 +48,15 @@ pub mod prelude {
     pub use smallvec::{self, SmallVec};
 
     pub use super::{
-        AccountHash, BackendEvent, BackendEventConsumer, BackendMailbox, BackendOp,
-        EnvelopeHashBatch, FlagOp, IsSubscribedFn, LazyCountSet, MailBackend,
-        MailBackendCapabilities, MailBackendExtensionStatus, Mailbox, MailboxHash,
-        MailboxPermissions, ReadOnlyOp, RefreshEvent, RefreshEventKind, TagHash,
+        AccountHash, BackendEvent, BackendEventConsumer, BackendMailbox, EnvelopeHashBatch, FlagOp,
+        IsSubscribedFn, LazyCountSet, MailBackend, MailBackendCapabilities,
+        MailBackendExtensionStatus, Mailbox, MailboxHash, MailboxPermissions, RefreshEvent,
+        RefreshEventKind, TagHash,
     };
     pub use crate::{
         conf::AccountSettings,
         email::{Envelope, EnvelopeHash, Flag},
-        error::{Error, ErrorKind, Result},
+        error::{Error, ErrorKind, NetworkErrorKind, Result},
         search::Query,
         Collection, HeaderName, LogLevel, SpecialUsageMailbox,
     };
@@ -297,6 +298,30 @@ impl From<Error> for BackendEvent {
     }
 }
 
+impl From<RefreshEvent> for BackendEvent {
+    /// Convert into [`BackendEvent::Refresh`] variant.
+    fn from(val: RefreshEvent) -> Self {
+        Self::Refresh(val)
+    }
+}
+
+impl TryFrom<Vec<RefreshEvent>> for BackendEvent {
+    type Error = ();
+
+    /// Convert a non-empty vector into [`BackendEvent::Refresh`] or
+    /// [`BackendEvent::RefreshBatch`] variants.
+    fn try_from(mut val: Vec<RefreshEvent>) -> std::result::Result<Self, Self::Error> {
+        if val.is_empty() {
+            return Err(());
+        }
+        if val.len() == 1 {
+            Ok(Self::Refresh(val.pop().unwrap()))
+        } else {
+            Ok(Self::RefreshBatch(val))
+        }
+    }
+}
+
 #[derive(Clone, Debug)]
 pub enum RefreshEventKind {
     Update(EnvelopeHash, Box<Envelope>),
@@ -423,9 +448,9 @@ pub trait MailBackend: ::std::fmt::Debug + Send + Sync {
     }
     fn fetch(&mut self, mailbox_hash: MailboxHash) -> ResultStream<Vec<Envelope>>;
     fn refresh(&mut self, mailbox_hash: MailboxHash) -> ResultFuture<()>;
-    fn watch(&self) -> ResultFuture<()>;
+    fn watch(&self) -> ResultStream<BackendEvent>;
     fn mailboxes(&self) -> ResultFuture<HashMap<MailboxHash, Mailbox>>;
-    fn operation(&self, hash: EnvelopeHash) -> Result<Box<dyn BackendOp>>;
+    fn envelope_bytes_by_hash(&self, hash: EnvelopeHash) -> ResultFuture<Vec<u8>>;
 
     fn save(
         &self,
@@ -538,65 +563,6 @@ pub trait MailBackend: ::std::fmt::Debug + Send + Sync {
     }
 }
 
-/// A `BackendOp` manages common operations for the various mail backends. They
-/// only live for the duration of the operation. They are generated by the
-/// `operation` method of `Mailbackend` trait.
-///
-/// # Motivation
-///
-/// We need a way to do various operations on individual mails regardless of
-/// what backend they come from (eg local or imap).
-///
-/// # Creation
-/// ```ignore
-/// /* Create operation from Backend */
-///
-/// let op = backend.operation(message.hash(), mailbox.hash());
-/// ```
-///
-/// # Example
-/// ```ignore
-/// use melib::backends::{BackendOp};
-/// use melib::Result;
-/// use melib::{Envelope, Flag};
-///
-/// #[derive(Debug)]
-/// struct FooOp {}
-///
-/// impl BackendOp for FooOp {
-///     fn as_bytes(&mut self) -> Result<&[u8]> {
-///         unimplemented!()
-///     }
-/// }
-///
-/// let operation = Box::new(FooOp {});
-/// ```
-pub trait BackendOp: ::std::fmt::Debug + ::std::marker::Send + ::std::marker::Sync {
-    fn as_bytes(&self) -> ResultFuture<Vec<u8>>;
-}
-
-/// Wrapper for [`BackendOp`] that are to be set read-only.
-///
-/// Warning: Backend implementations may still cause side-effects (for example
-/// IMAP can set the Seen flag when fetching an envelope)
-#[derive(Debug)]
-pub struct ReadOnlyOp {
-    op: Box<dyn BackendOp>,
-}
-
-impl ReadOnlyOp {
-    #[allow(clippy::new_ret_no_self)]
-    pub fn new(op: Box<dyn BackendOp>) -> Box<dyn BackendOp> {
-        Box::new(Self { op })
-    }
-}
-
-impl BackendOp for ReadOnlyOp {
-    fn as_bytes(&self) -> ResultFuture<Vec<u8>> {
-        self.op.as_bytes()
-    }
-}
-
 #[derive(Clone, Copy, Debug, Default, Deserialize, Eq, Hash, PartialEq, Serialize)]
 pub enum SpecialUsageMailbox {
     #[default]
@@ -729,9 +695,10 @@ impl From<EnvelopeHash> for EnvelopeHashBatch {
     }
 }
 
-impl std::convert::TryFrom<&[EnvelopeHash]> for EnvelopeHashBatch {
+impl TryFrom<&[EnvelopeHash]> for EnvelopeHashBatch {
     type Error = ();
 
+    /// Convert a non-empty [`EnvelopeHash`] array.
     fn try_from(value: &[EnvelopeHash]) -> std::result::Result<Self, Self::Error> {
         if value.is_empty() {
             return Err(());
