@@ -23,7 +23,7 @@
 #![allow(clippy::type_complexity)]
 
 #[cfg(test)]
-mod tests;
+pub mod tests;
 
 #[cfg(any(test, doc))]
 use std::backtrace::Backtrace;
@@ -212,63 +212,33 @@ impl<I> nom::error::ContextError<I> for ParsingError<I> {}
 
 impl<'i> From<ParsingError<&'i [u8]>> for Error {
     fn from(val: ParsingError<&'i [u8]>) -> Self {
-        Self::new("Parsing error")
-            .set_summary(format!(
-                r#"In input: "{}...",
-Error: {}"#,
-                String::from_utf8_lossy(val.input)
-                    .chars()
-                    .take(30)
-                    .collect::<String>(),
-                val.error
-            ))
-            .set_details({
-                #[cfg(any(test, doc))]
-                {
-                    println!(
-                        "\tInput:\n{}\tError:\n{}\n\tBacktrace:\n{}",
-                        String::from_utf8_lossy(val.input)
-                            .chars()
-                            .take(30)
-                            .collect::<String>(),
-                        val.error,
-                        val.backtrace
-                    );
-                    val.backtrace.to_string()
-                }
-                #[cfg(not(any(test, doc)))]
-                {
-                    ""
-                }
-            })
+        #[cfg(any(test, doc))]
+        {
+            eprintln!(
+                "Parsing error for input:\n{}\nError:\n{}\nBacktrace:\n{}",
+                String::from_utf8_lossy(val.input),
+                val.error,
+                val.backtrace
+            );
+        }
+        Self::new(format!(
+            "Error when parsing: \"{}\"",
+            String::from_utf8_lossy(val.input)
+        ))
+        .set_details(val.error)
     }
 }
 
 impl<'i> From<ParsingError<&'i str>> for Error {
     fn from(val: ParsingError<&'i str>) -> Self {
-        Self::new("Parsing error")
-            .set_summary(format!(
-                r#"In input: "{}...",
-Error: {}"#,
-                val.input.chars().take(30).collect::<String>(),
-                val.error
-            ))
-            .set_details({
-                #[cfg(any(test, doc))]
-                {
-                    println!(
-                        "\tInput:\n{}\tError:\n{}\n\tBacktrace:\n{}",
-                        val.input.chars().take(30).collect::<String>(),
-                        val.error,
-                        val.backtrace
-                    );
-                    val.backtrace.to_string()
-                }
-                #[cfg(not(any(test, doc)))]
-                {
-                    ""
-                }
-            })
+        #[cfg(any(test, doc))]
+        {
+            eprintln!(
+                "Parsing error for input:\n{}\nError:\n{}\nBacktrace:\n{}",
+                val.input, val.error, val.backtrace
+            );
+        }
+        Self::new(format!("Error when parsing: \"{}\"", val.input)).set_details(val.error)
     }
 }
 
@@ -1036,11 +1006,14 @@ pub mod generic {
         let mut headers = HeaderMap::empty();
         let mut address: Vec<Address>;
 
-        if String::from_utf8_lossy(input).matches('?').count() > 1 {
-            return Err(nom::Err::Error(
-                (input, "mailto(): Using '?' twice is invalid.").into(),
-            ));
-        }
+        // [ref:TODO]: Validate header content (e.g. test address containing Headers for valid address,
+        // so that the following would return an error:
+
+        // if String::from_utf8_lossy(input).matches('?').count() > 1 {
+        //     return Err(nom::Err::Error(
+        //         (input, "mailto(): Using '?' twice is invalid.").into(),
+        //     ));
+        // }
         input = &input[b"mailto:".len()..];
         let Ok(mut decoded_owned) = String::from_utf8(input.to_vec()) else {
             return Err(nom::Err::Error(
@@ -1067,7 +1040,7 @@ pub mod generic {
         let end = decoded.as_bytes().iter().position(|e| *e == b'?');
         let end_or_len = end.unwrap_or(decoded.len());
 
-        if let Ok(addr) = percent_decode(decoded[..end_or_len].as_bytes())
+        if let Ok(addr) = percent_decode(&decoded.as_bytes()[..end_or_len])
             .decode_utf8()
             .map_err(|_| nom::Err::Error((input, "mailto(): Not valid UTF-8.")))
             .and_then(|s| {
@@ -1077,7 +1050,7 @@ pub mod generic {
             })
         {
             address = addr;
-            decoded = if decoded[end_or_len..].is_empty() {
+            decoded = if decoded.as_bytes()[end_or_len..].is_empty() {
                 &decoded[end_or_len..]
             } else {
                 &decoded[end_or_len + 1..]
@@ -1110,13 +1083,13 @@ pub mod generic {
 
         let mut i = 0;
         while !decoded[i..].is_empty() {
-            let tag = if let Some(tag_pos) = decoded[i..].as_bytes().iter().position(|e| *e == b'=')
+            let tag = if let Some(tag_pos) = decoded.as_bytes()[i..].iter().position(|e| *e == b'=')
             {
-                let ret = &decoded[i..][0..tag_pos];
+                let ret = &decoded.as_bytes()[i..][0..tag_pos];
                 i += tag_pos + 1;
                 ret
-            } else if decoded[i..].as_bytes().starts_with(b"body") {
-                let ret = &decoded[i..][0.."body".len()];
+            } else if decoded.as_bytes()[i..].starts_with(b"body") {
+                let ret = &decoded.as_bytes()[i..][0.."body".len()];
                 i += "body".len() + 1;
                 ret
             } else {
@@ -1132,20 +1105,19 @@ pub mod generic {
                 ));
             };
 
-            let value_end = decoded[i..]
-                .as_bytes()
+            let value_end = decoded.as_bytes()[i..]
                 .iter()
                 .position(|e| *e == b'&')
-                .unwrap_or_else(|| decoded[i..].len());
+                .unwrap_or_else(|| decoded.as_bytes()[i..].len());
 
-            let Ok(value) = percent_decode(decoded[i..][..value_end].as_bytes())
+            let Ok(value) = percent_decode(&decoded.as_bytes()[i..][..value_end])
                 .decode_utf8()
                 .map(|v| v.to_string())
             else {
                 return Err(nom::Err::Error((input, "mailto(): invalid UTF-8.").into()));
             };
             match tag {
-                "body" if body.is_none() => {
+                b"body" if body.is_none() => {
                     body = Some(value);
                 }
                 other => match HeaderName::try_from(other) {
@@ -1184,7 +1156,7 @@ pub mod generic {
                     }
                 },
             }
-            if decoded[i..][value_end..].is_empty() {
+            if decoded.as_bytes()[i..][value_end..].is_empty() {
                 break;
             }
             i += value_end + 1;
@@ -2191,6 +2163,7 @@ pub mod encodings {
             Charset::UTF16 => Ok(UTF_16LE.decode(s).0.to_string()),
             Charset::ISO2022JP => Ok(ISO_2022_JP.decode(s).0.to_string()),
             Charset::EUCJP => Ok(EUC_JP.decode(s).0.to_string()),
+            Charset::KSX1001 => Ok(EUC_KR.decode(s).0.to_string()),
         }
     }
 
@@ -2276,12 +2249,14 @@ pub mod encodings {
         let mut ptr = 0;
 
         while ptr < input.len() {
+            let mut token_start = ptr;
             let mut flag = false;
             // Check if word is encoded.
             while let Ok((rest, v)) = encoded_word(&input[ptr..]) {
                 flag = true;
                 input = rest;
                 ptr = 0;
+                token_start = 0;
                 acc.extend(v);
 
                 // consume whitespace
@@ -2293,8 +2268,8 @@ pub mod encodings {
                     break;
                 }
             }
-            if flag && ptr < input.len() && ptr != 0 {
-                acc.push(b' ');
+            if flag && token_start != ptr {
+                acc.extend(ascii_token(&input[token_start..ptr])?.1);
             }
             let end = input[ptr..].find(b"=?");
 
@@ -2625,7 +2600,39 @@ pub mod address {
 
     /// `address         =   mailbox / group`
     pub fn address(input: &[u8]) -> IResult<&[u8], Address> {
-        alt((mailbox, group))(input.ltrim())
+        let err = match alt((mailbox, group))(input.ltrim()) {
+            ok @ Ok(_) => return ok,
+            Err(err) => err,
+        };
+        // Try to detect common errors
+        static RE: std::sync::OnceLock<regex::bytes::Regex> = std::sync::OnceLock::new();
+
+        let re = RE.get_or_init(|| {
+            regex::bytes::Regex::new(r"(?m)(?:^(?P<display>.*?) [<](?P<addr_spec>.*?[@].*?)[>]$)|(?:^()(?P<addr_spec_bare>[^<@]*?[@][^>]*$))|(?:^()[<](?P<addr_spec_bare_bracketed>.*?[@].*?)[>]$)").unwrap()
+        });
+
+        if let Some((full, [display_name, addr_spec])) =
+            re.captures(input).map(|cap| cap.extract::<2>())
+        {
+            if [
+                b'(', b')', b'<', b'>', b'[', b']', b':', b';', b'@', b'\\', b',', b'.', b'"',
+            ]
+            .iter()
+            .any(|b| display_name.contains(b))
+            {
+                return Err(nom::Err::Error(ParsingError::<&[u8]>::new(
+                    full,
+                    Cow::Owned(format!(
+                        "Address `{}` contains errors: Display names that contain ()<>[]:;@\\,.\" \
+                         must be quoted. Try: `\"{}\" <{}>` instead",
+                        String::from_utf8_lossy(input),
+                        String::from_utf8_lossy(display_name),
+                        String::from_utf8_lossy(addr_spec)
+                    )),
+                )));
+            }
+        }
+        Err(err)
     }
 
     pub fn rfc2822address_list(input: &[u8]) -> IResult<&[u8], SmallVec<[Address; 1]>> {
