@@ -24,17 +24,24 @@ use std::{
     convert::TryFrom,
     future::Future,
     pin::Pin,
+    str::FromStr,
     sync::Arc,
     time::{Duration, Instant, SystemTime},
 };
 
-use futures::io::{AsyncReadExt, AsyncWriteExt};
+use futures::{
+    io::{AsyncReadExt, AsyncWriteExt},
+    lock::{
+        MappedMutexGuard as FutureMappedMutexGuard, Mutex as FutureMutex,
+        MutexGuard as FutureMutexGuard,
+    },
+};
 use imap_codec::{
     encode::{Encoder, Fragment},
     imap_types::{
         auth::AuthMechanism,
         command::{Command, CommandBody},
-        core::{AString, LiteralMode, Tag, Vec1},
+        core::{AString, Literal, LiteralMode, Tag, Vec1},
         extensions::{compress::CompressionAlgorithm, enable::CapabilityEnable},
         mailbox::Mailbox,
         search::SearchKey,
@@ -49,7 +56,10 @@ use native_tls::TlsConnector;
 pub use smol::Async as AsyncWrapper;
 
 use crate::{
-    backends::{BackendEvent, MailboxHash, RefreshEvent},
+    backends::{
+        prelude::{EnvelopeHash, Flag, Query},
+        BackendEvent, BackendMailbox, MailboxHash, RefreshEvent,
+    },
     email::parser::BytesExt,
     error::*,
     imap::{
@@ -57,7 +67,8 @@ use crate::{
             self, id_ext::id_ext_response, ImapLineSplit, ImapResponse, RequiredResponses,
             ResponseCode, SelectResponse,
         },
-        Capabilities, ImapServerConf, UIDStore,
+        search::ToImapSearch,
+        Capabilities, ImapServerConf, UIDStore, UID,
     },
     text::Truncate,
     utils::{
@@ -95,6 +106,14 @@ pub enum ImapProtocol {
     ManageSieve,
 }
 
+impl Default for ImapProtocol {
+    fn default() -> Self {
+        Self::IMAP {
+            extension_use: ImapExtensionUse::default(),
+        }
+    }
+}
+
 impl ImapProtocol {
     /// If `self` variant supports the `DEFLATE` extension, set a new value for
     /// whether it will be used or not, otherwise silently ignore the new
@@ -184,6 +203,163 @@ async fn try_await(cl: impl Future<Output = Result<()>> + Send) -> Result<()> {
     cl.await
 }
 
+const POOL_LIMIT: usize = 5;
+
+#[derive(Debug)]
+pub struct ConnectionMutex {
+    pub inner: FutureMutex<ImapConnection>,
+    pub pool: [FutureMutex<Option<ImapConnection>>; POOL_LIMIT],
+    pub server_conf: ImapServerConf,
+    pub uid_store: Arc<UIDStore>,
+    pub use_connection_pool: bool,
+}
+
+#[derive(Debug)]
+pub enum ConnectionMutexGuard<'a> {
+    Main(FutureMutexGuard<'a, ImapConnection>),
+    Pool(FutureMappedMutexGuard<'a, Option<ImapConnection>, ImapConnection>),
+}
+
+impl std::ops::Deref for ConnectionMutexGuard<'_> {
+    type Target = ImapConnection;
+
+    fn deref(&self) -> &Self::Target {
+        match self {
+            Self::Main(inner) => std::ops::Deref::deref(inner),
+            Self::Pool(inner) => std::ops::Deref::deref(inner),
+        }
+    }
+}
+
+impl std::ops::DerefMut for ConnectionMutexGuard<'_> {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        match self {
+            Self::Main(inner) => std::ops::DerefMut::deref_mut(inner),
+            Self::Pool(inner) => std::ops::DerefMut::deref_mut(inner),
+        }
+    }
+}
+
+impl ConnectionMutex {
+    #[inline]
+    pub fn new(
+        inner: ImapConnection,
+        server_conf: ImapServerConf,
+        uid_store: Arc<UIDStore>,
+        use_connection_pool: bool,
+    ) -> Self {
+        let inner = inner.into();
+
+        Self {
+            inner,
+            pool: [
+                None.into(),
+                None.into(),
+                None.into(),
+                None.into(),
+                None.into(),
+            ],
+            server_conf,
+            uid_store,
+            use_connection_pool,
+        }
+    }
+
+    #[inline]
+    pub async fn lock(&self) -> Result<ConnectionMutexGuard<'_>> {
+        let timeout_dur = self.uid_store.timeout;
+        if !self.use_connection_pool {
+            return Ok(ConnectionMutexGuard::Main(
+                timeout(timeout_dur, self.inner.lock()).await?,
+            ));
+        }
+        // Fast check if main connection is available.
+        if let Some(guard) = self.inner.try_lock() {
+            return Ok(ConnectionMutexGuard::Main(guard));
+        }
+
+        async fn initialize_pool_conn_fn(
+            i: usize,
+            mutex: &ConnectionMutex,
+        ) -> Result<ConnectionMutexGuard<'_>> {
+            if let Some(dur) = mutex.uid_store.timeout {
+                // Only begin potentially initializing a new connection if the main connection
+                // has not been immediately available.
+                smol::Timer::after(dur / 4).await;
+            }
+            let mut guard = timeout(mutex.uid_store.timeout, mutex.pool[i].lock()).await?;
+            if guard.is_none() {
+                let mut new_pool_entry = ImapConnection::new_connection(
+                    &mutex.server_conf,
+                    format!("imap-connection-pool-{}", i).into(),
+                    mutex.uid_store.clone(),
+                    false,
+                );
+                if let Err(err) = timeout(mutex.uid_store.timeout, new_pool_entry.connect()).await?
+                {
+                    log::trace!(
+                        "{} Creation of new connection for connection pool failed: {}",
+                        mutex.uid_store.account_name,
+                        err
+                    );
+                    return Err(err);
+                }
+                *guard = Some(new_pool_entry);
+                log::trace!(
+                    "{} Created new pool connection #{}",
+                    mutex.uid_store.account_name,
+                    i,
+                );
+            }
+            // At this point, `guard` points to a `Some(_)` value.
+            Ok(ConnectionMutexGuard::Pool(FutureMutexGuard::map(
+                guard,
+                |o: &mut Option<ImapConnection>| {
+                    // SAFETY: we checked if `guard` is None above and initialized it if it
+                    // was.
+                    unsafe { o.as_mut().unwrap_unchecked() }
+                },
+            )))
+        }
+
+        let main_fut = async {
+            timeout(timeout_dur, self.inner.lock())
+                .await
+                .map(ConnectionMutexGuard::Main)
+        };
+
+        let (f0, f1, f2, f3, f4) = (
+            initialize_pool_conn_fn(0, self),
+            initialize_pool_conn_fn(1, self),
+            initialize_pool_conn_fn(2, self),
+            initialize_pool_conn_fn(3, self),
+            initialize_pool_conn_fn(4, self),
+        );
+        futures::pin_mut!(main_fut);
+        futures::pin_mut!(f0);
+        futures::pin_mut!(f1);
+        futures::pin_mut!(f2);
+        futures::pin_mut!(f3);
+        futures::pin_mut!(f4);
+        match futures::future::select(futures::future::select_ok([f0, f1, f2, f3, f4]), main_fut)
+            .await
+        {
+            // The main connection was the first available
+            futures::future::Either::Right((Ok(main), _)) => Ok(main),
+            // The main connection timed out, so fallback to waiting for a pool connection
+            futures::future::Either::Right((Err(_), pool_fut)) => {
+                let (res, _) = pool_fut.await?;
+                Ok(res)
+            }
+            // A pool conn was the first available
+            futures::future::Either::Left((Ok((res, _)), _main)) => Ok(res),
+            // All pool connections were unavailable, so fallback to waiting for the main
+            // connection
+            futures::future::Either::Left((Err(_), main_fut)) => main_fut.await,
+        }
+    }
+}
+
 #[derive(Debug)]
 pub struct ImapConnection {
     pub id: Cow<'static, str>,
@@ -191,6 +367,7 @@ pub struct ImapConnection {
     pub server_conf: ImapServerConf,
     pub sync_policy: SyncPolicy,
     pub uid_store: Arc<UIDStore>,
+    pub send_state_changes: bool,
 }
 
 impl ImapStream {
@@ -198,17 +375,19 @@ impl ImapStream {
         server_conf: &ImapServerConf,
         id: Cow<'static, str>,
         uid_store: &UIDStore,
+        send_state_changes: bool,
     ) -> Result<(Capabilities, Self)> {
         let path = &server_conf.server_hostname;
 
-        let cmd_id = 1;
         let stream = if server_conf.use_tls {
-            (uid_store.event_consumer)(
-                uid_store.account_hash,
-                BackendEvent::AccountStateChange {
-                    message: "Establishing TLS connection.".into(),
-                },
-            );
+            if send_state_changes {
+                (uid_store.event_consumer)(
+                    uid_store.account_hash,
+                    BackendEvent::AccountStateChange {
+                        message: "Establishing TLS connection.".into(),
+                    },
+                );
+            }
             let mut connector = TlsConnector::builder();
             if server_conf.danger_accept_invalid_certs {
                 connector.danger_accept_invalid_certs(true);
@@ -244,7 +423,7 @@ impl ImapStream {
                 let mut buf = vec![0; Connection::IO_BUF_SIZE];
                 match server_conf.protocol {
                     ImapProtocol::IMAP { .. } => socket
-                        .write_all(format!("M{} STARTTLS\r\n", cmd_id).as_bytes())
+                        .write_all(b"M1 STARTTLS\r\n")
                         .await
                         .chain_err_summary(err_fn)?,
                     ImapProtocol::ManageSieve => {
@@ -349,13 +528,13 @@ impl ImapStream {
         };
         if let Err(err) = stream
             .get_ref()
-            .set_keepalive(Some(Duration::new(60 * 9, 0)))
+            .set_keepalive(Some(Duration::new(60 * 35, 0)))
         {
             log::warn!("Could not set TCP keepalive in IMAP connection: {}", err);
         }
         let mut res = Vec::with_capacity(8 * 1024);
         let mut ret = Self {
-            cmd_id,
+            cmd_id: 1,
             id,
             stream,
             protocol: server_conf.protocol,
@@ -377,12 +556,14 @@ impl ImapStream {
             return Ok((Default::default(), ret));
         }
 
-        (uid_store.event_consumer)(
-            uid_store.account_hash,
-            BackendEvent::AccountStateChange {
-                message: "Negotiating server capabilities.".into(),
-            },
-        );
+        if send_state_changes {
+            (uid_store.event_consumer)(
+                uid_store.account_hash,
+                BackendEvent::AccountStateChange {
+                    message: "Negotiating server capabilities.".into(),
+                },
+            );
+        }
         ret.send_command(CommandBody::Capability).await?;
         ret.read_response(&mut res).await?;
 
@@ -439,12 +620,14 @@ impl ImapStream {
             .set_kind(ErrorKind::ProtocolNotSupported));
         }
 
-        (uid_store.event_consumer)(
-            uid_store.account_hash,
-            BackendEvent::AccountStateChange {
-                message: "Attempting authentication.".into(),
-            },
-        );
+        if send_state_changes {
+            (uid_store.event_consumer)(
+                uid_store.account_hash,
+                BackendEvent::AccountStateChange {
+                    message: "Attempting authentication.".into(),
+                },
+            );
+        }
         match server_conf.protocol {
             ImapProtocol::IMAP {
                 extension_use: ImapExtensionUse { oauth2, .. },
@@ -775,6 +958,7 @@ impl ImapConnection {
         server_conf: &ImapServerConf,
         id: Cow<'static, str>,
         uid_store: Arc<UIDStore>,
+        send_state_changes: bool,
     ) -> Self {
         Self {
             stream: Err(Error::new("Offline".to_string())),
@@ -789,6 +973,7 @@ impl ImapConnection {
                 SyncPolicy::None
             },
             uid_store,
+            send_state_changes,
         }
     }
 
@@ -832,9 +1017,13 @@ impl ImapConnection {
                     return Ok(());
                 }
             }
-            let new_stream =
-                ImapStream::new_connection(&self.server_conf, self.id.clone(), &self.uid_store)
-                    .await;
+            let new_stream = ImapStream::new_connection(
+                &self.server_conf,
+                self.id.clone(),
+                &self.uid_store,
+                self.send_state_changes,
+            )
+            .await;
             if let Err(err) = new_stream.as_ref() {
                 self.uid_store.is_online.lock().unwrap().1 = Err(err.clone());
             } else {
@@ -973,7 +1162,7 @@ impl ImapConnection {
                             return r.into();
                         }
                         ImapResponse::No(ref _response_code)
-                            if required_responses.intersects(RequiredResponses::NO_REQUIRED) =>
+                            if required_responses.intersects(RequiredResponses::NO) =>
                         {
                             imap_log!(
                                 trace,
@@ -1023,13 +1212,16 @@ impl ImapConnection {
                         }
                         _ => {}
                     }
-                    /* imap_log!(trace, self,
-                        "check every line for required_responses: {:#?}",
-                        &required_responses
-                    );*/
                     for l in response.split_rn() {
-                        /* imap_log!(trace, self, "check line: {}", &l); */
-                        if required_responses.check(l) || !self.process_untagged(l).await? {
+                        if required_responses.check(l) {
+                            ret.extend_from_slice(l);
+                        } else if let Ok(Some(untagged_response)) =
+                            super::protocol_parser::untagged_responses(l).map(|(_, v, _)| v)
+                        {
+                            if let Some(ev) = self.process_untagged(untagged_response).await? {
+                                self.add_backend_event(ev);
+                            }
+                        } else {
                             ret.extend_from_slice(l);
                         }
                     }
@@ -1153,8 +1345,7 @@ impl ImapConnection {
         }
         self.send_command(CommandBody::select(imap_path.as_str())?)
             .await?;
-        self.read_response(ret, RequiredResponses::SELECT_REQUIRED)
-            .await?;
+        self.read_response(ret, RequiredResponses::SELECT).await?;
         imap_log!(
             trace,
             self,
@@ -1231,8 +1422,7 @@ impl ImapConnection {
         }
         self.send_command(CommandBody::examine(imap_path.as_str())?)
             .await?;
-        self.read_response(ret, RequiredResponses::EXAMINE_REQUIRED)
-            .await?;
+        self.read_response(ret, RequiredResponses::EXAMINE).await?;
 
         imap_log!(
             trace,
@@ -1290,7 +1480,7 @@ impl ImapConnection {
                         }
                     }
                     self.send_command(CommandBody::select(nonexistent)?).await?;
-                    self.read_response(&mut response, RequiredResponses::NO_REQUIRED)
+                    self.read_response(&mut response, RequiredResponses::NO)
                         .await?;
                 }
             }
@@ -1421,3 +1611,123 @@ async fn read(
     }
     None
 }
+
+/// Async methods counterparts of the [`MailBackend`](crate::MailBackend) crate.
+/// Direct use should minimise cloning, locks and delay compared to using
+/// `MailBackend` methods.
+impl ImapConnection {
+    pub async fn refresh(&mut self, mailbox_hash: MailboxHash) -> Result<()> {
+        let mailbox = timeout(self.uid_store.timeout, self.uid_store.mailboxes.lock())
+            .await?
+            .get(&mailbox_hash)
+            .cloned()
+            .unwrap();
+        if let Some(ev) = crate::imap::watch::examine_updates(mailbox, self).await? {
+            self.add_backend_event(ev);
+        }
+        Ok(())
+    }
+
+    pub async fn is_online(&mut self) -> Result<()> {
+        imap_log!(trace, self, "is_online");
+        let timeout_dur = self.uid_store.timeout;
+        match timeout(timeout_dur, self.connect()).await {
+            Ok(Ok(())) => Ok(()),
+            Err(err) | Ok(Err(err)) => {
+                self.stream = Err(err.clone());
+                self.connect().await
+            }
+        }
+    }
+
+    pub async fn save(
+        &mut self,
+        bytes: Vec<u8>,
+        mailbox_hash: MailboxHash,
+        flags: Option<Flag>,
+    ) -> Result<()> {
+        let mut response = Vec::with_capacity(8 * 1024);
+        self.select_mailbox(mailbox_hash, &mut response, true)
+            .await?;
+        let path = {
+            let mailboxes = self.uid_store.mailboxes.lock().await;
+
+            let mailbox = mailboxes.get(&mailbox_hash).ok_or_else(|| {
+                Error::new(format!("Mailbox with hash {} not found.", mailbox_hash))
+            })?;
+            if !mailbox.permissions.lock().unwrap().create_messages {
+                return Err(Error::new(format!(
+                    "You are not allowed to create messages in mailbox {}",
+                    mailbox.path()
+                )));
+            }
+
+            mailbox.imap_path().to_string()
+        };
+        let flags = flags.unwrap_or_else(Flag::empty);
+        let data = {
+            let capabilities = self.uid_store.capabilities.lock().unwrap();
+            let has_literal_plus: bool = capabilities
+                .iter()
+                .any(|cap| cap.eq_ignore_ascii_case(b"LITERAL+"));
+            let has_literal_minus: bool = capabilities
+                .iter()
+                .any(|cap| cap.eq_ignore_ascii_case(b"LITERAL-"));
+            drop(capabilities);
+            if has_literal_plus || (has_literal_minus && bytes.len() <= 4096) {
+                Literal::try_from(bytes)?.into_non_sync()
+            } else {
+                Literal::try_from(bytes)?
+            }
+        };
+        self.send_command(CommandBody::append(path, flags.into(), None, data)?)
+            .await?;
+        // [ref:TODO]: check for APPENDUID [RFC4315 - UIDPLUS]
+        self.read_response(&mut response, RequiredResponses::empty())
+            .await?;
+        Ok(())
+    }
+
+    pub async fn search(
+        &mut self,
+        query: Query,
+        mailbox_hash: Option<MailboxHash>,
+    ) -> Result<Vec<EnvelopeHash>> {
+        let Some(mailbox_hash) = mailbox_hash else {
+            return Err(Error::new(
+                "Cannot search without specifying mailbox on IMAP",
+            ));
+        };
+        let query_str = query.to_imap_search();
+
+        let mut response = Vec::with_capacity(8 * 1024);
+        self.examine_mailbox(mailbox_hash, &mut response, false)
+            .await?;
+        self.send_command_raw(format!("UID SEARCH CHARSET UTF-8 {}", query_str.trim()).as_bytes())
+            .await?;
+        self.read_response(&mut response, RequiredResponses::SEARCH)
+            .await?;
+        imap_log!(
+            trace,
+            self,
+            "searching for {} returned: {}",
+            query_str,
+            String::from_utf8_lossy(&response)
+        );
+
+        for l in response.split_rn() {
+            if l.starts_with(b"* SEARCH") {
+                let uid_index = self.uid_store.uid_index.lock()?;
+                return Ok(Vec::from_iter(
+                    String::from_utf8_lossy(l[b"* SEARCH".len()..].trim())
+                        .split_whitespace()
+                        .map(UID::from_str)
+                        .filter_map(std::result::Result::ok)
+                        .filter_map(|uid| uid_index.get(&(mailbox_hash, uid)))
+                        .copied(),
+                ));
+            }
+        }
+        Err(Error::new(String::from_utf8_lossy(&response).to_string()))
+    }
+}
