@@ -39,7 +39,7 @@ pub enum FetchStage {
 #[derive(Debug)]
 pub struct FetchState {
     pub stage: FetchStage,
-    pub connection: Arc<FutureMutex<ImapConnection>>,
+    pub connection: Arc<ConnectionMutex>,
     pub mailbox_hash: MailboxHash,
     pub uid_store: Arc<UIDStore>,
     pub batch_size: usize,
@@ -55,7 +55,7 @@ impl FetchState {
                     let select_response = self
                         .connection
                         .lock()
-                        .await
+                        .await?
                         .init_mailbox(self.mailbox_hash)
                         .await?;
                     _ = self
@@ -75,7 +75,7 @@ impl FetchState {
                     let select_response = self
                         .connection
                         .lock()
-                        .await
+                        .await?
                         .init_mailbox(self.mailbox_hash)
                         .await?;
                     if let Err(err) = self
@@ -98,7 +98,7 @@ impl FetchState {
                         Err(err) => {
                             imap_log!(
                                 error,
-                                self.connection.lock().await,
+                                self.connection.lock().await?,
                                 "IMAP cache error: could not fetch cache for {}. Reason: {}",
                                 self.uid_store.account_name,
                                 err
@@ -107,7 +107,7 @@ impl FetchState {
                             if let Err(err) = self.uid_store.reset() {
                                 imap_log!(
                                     error,
-                                    self.connection.lock().await,
+                                    self.connection.lock().await?,
                                     "IMAP cache error: could not reset cache for {}. Reason: {}",
                                     self.uid_store.account_name,
                                     err
@@ -157,7 +157,7 @@ impl FetchState {
                         Err(err) => {
                             imap_log!(
                                 error,
-                                self.connection.lock().await,
+                                self.connection.lock().await?,
                                 "IMAP cache error: could not fetch cache for {}. Reason: {}",
                                 self.uid_store.account_name,
                                 err
@@ -166,7 +166,7 @@ impl FetchState {
                             if let Err(err) = self.uid_store.reset() {
                                 imap_log!(
                                     error,
-                                    self.connection.lock().await,
+                                    self.connection.lock().await?,
                                     "IMAP cache error: could not reset cache for {}. Reason: {}",
                                     self.uid_store.account_name,
                                     err
@@ -182,7 +182,7 @@ impl FetchState {
                     }
                 }
                 FetchStage::ResyncCache => {
-                    let mut conn = self.connection.lock().await;
+                    let mut conn = self.connection.lock().await?;
                     let select_response = conn.init_mailbox(self.mailbox_hash).await?;
                     match self
                         .uid_store
@@ -240,7 +240,7 @@ impl FetchState {
                         self.stage = FetchStage::Finished;
                         return Ok(Vec::new());
                     }
-                    let mut conn = connection.lock().await;
+                    let mut conn = connection.lock().await?;
                     let mut response = Vec::with_capacity(8 * 1024);
                     let max_uid_left = max_uid;
 
@@ -256,13 +256,15 @@ impl FetchState {
 
                             SequenceSet::try_from(min..=max)?
                         };
+                        let (required_responses, macro_or_item_names) =
+                            crate::imap::email::common_attributes();
                         conn.send_command(CommandBody::Fetch {
                             sequence_set,
-                            macro_or_item_names: common_attributes(),
+                            macro_or_item_names,
                             uid: true,
                         })
                         .await?;
-                        conn.read_response(&mut response, RequiredResponses::FETCH_REQUIRED)
+                        conn.read_response(&mut response, required_responses)
                             .await
                             .chain_err_summary(|| {
                                 format!(
@@ -296,7 +298,16 @@ impl FetchState {
                                     "response was: {}",
                                     String::from_utf8_lossy(&response)
                                 );
-                                conn.process_untagged(raw_fetch_value).await?;
+                                if let Ok(Some(untagged_response)) =
+                                    super::protocol_parser::untagged_responses(raw_fetch_value)
+                                        .map(|(_, v, _)| v)
+                                {
+                                    if let Some(ev) =
+                                        conn.process_untagged(untagged_response).await?
+                                    {
+                                        conn.add_backend_event(ev);
+                                    }
+                                }
                                 continue;
                             }
                             let uid = uid.unwrap();
@@ -429,7 +440,7 @@ impl FetchState {
             return Ok(None);
         }
         {
-            let mut conn = connection.lock().await;
+            let mut conn = connection.lock().await?;
             let select_response = conn.init_mailbox(mailbox_hash).await?;
             match Self::load_cache(
                 &conn,
