@@ -37,6 +37,7 @@ mod watch;
 pub use watch::*;
 mod search;
 pub use search::*;
+pub mod email;
 pub mod error;
 pub mod fetch;
 pub mod managesieve;
@@ -60,7 +61,7 @@ use fetch::{FetchStage, FetchState};
 pub extern crate imap_codec;
 use imap_codec::imap_types::{
     command::CommandBody,
-    core::{Atom, Literal},
+    core::Atom,
     flag::{Flag as ImapCodecFlag, StoreResponse, StoreType},
     mailbox::Mailbox as ImapTypesMailbox,
     sequence::{SequenceSet, ONE},
@@ -71,9 +72,10 @@ use crate::{
     backends::{prelude::*, RefreshEventKind::*},
     collection::Collection,
     conf::AccountSettings,
-    email::{parser::BytesExt, *},
+    email::*,
     error::{Error, ErrorKind, Result, ResultIntoError},
     imap::{protocol_parser::id_ext::IDResponse, sync::cache::ImapCache},
+    text::Truncate,
     utils::futures::timeout,
 };
 
@@ -95,6 +97,7 @@ pub static SUPPORTED_CAPABILITIES: &[&str] = &[
     "LIST-EXTENDED",
     "LIST-STATUS",
     "LITERAL+",
+    "LITERAL-",
     "LOGIN",
     "LOGINDISABLED",
     "MOVE",
@@ -155,6 +158,7 @@ macro_rules! get_conf_val {
 
 #[derive(Debug)]
 pub struct UIDStore {
+    pub is_subscribed: IsSubscribedFn,
     pub account_hash: AccountHash,
     pub account_name: Arc<str>,
     pub server_id: Arc<Mutex<Option<IDResponse>>>,
@@ -182,6 +186,7 @@ pub struct UIDStore {
 
 impl UIDStore {
     pub fn new(
+        is_subscribed: IsSubscribedFn,
         account_hash: AccountHash,
         account_name: Arc<str>,
         event_consumer: BackendEventConsumer,
@@ -191,6 +196,7 @@ impl UIDStore {
         Self {
             account_hash,
             account_name,
+            is_subscribed,
             server_id: Default::default(),
             keep_offline_cache: AtomicBool::new(keep_offline_cache),
             offline_cache: Arc::new(Mutex::new(None)),
@@ -217,7 +223,10 @@ impl UIDStore {
 
     fn init_cache(
         self: &Arc<Self>,
-        mutex: &mut std::sync::MutexGuard<'_, Option<Box<dyn ImapCache>>>,
+        #[allow(clippy::needless_pass_by_ref_mut)] mutex: &mut std::sync::MutexGuard<
+            '_,
+            Option<Box<dyn ImapCache>>,
+        >,
     ) -> Result<()> {
         if mutex.is_none() {
             #[cfg(feature = "sqlite3")]
@@ -240,8 +249,7 @@ impl UIDStore {
 
 #[derive(Debug)]
 pub struct ImapType {
-    pub _is_subscribed: IsSubscribedFn,
-    pub connection: Arc<FutureMutex<ImapConnection>>,
+    pub connection: Arc<ConnectionMutex>,
     pub server_conf: ImapServerConf,
     pub uid_store: Arc<UIDStore>,
 }
@@ -388,7 +396,7 @@ impl MailBackend for ImapType {
         };
 
         Ok(Box::pin(try_fn_stream(|emitter| async move {
-            let id = state.connection.lock().await.id.clone();
+            let id = state.connection.lock().await?.id.clone();
             {
                 let f = &state.uid_store.mailboxes.lock().await[&mailbox_hash];
                 f.set_warm(true);
@@ -427,16 +435,9 @@ impl MailBackend for ImapType {
 
     fn refresh(&mut self, mailbox_hash: MailboxHash) -> ResultFuture<()> {
         let main_conn = self.connection.clone();
-        let uid_store = self.uid_store.clone();
         Ok(Box::pin(async move {
-            let inbox = timeout(uid_store.timeout, uid_store.mailboxes.lock())
-                .await?
-                .get(&mailbox_hash)
-                .cloned()
-                .unwrap();
-            let mut conn = timeout(uid_store.timeout, main_conn.lock()).await?;
-            watch::examine_updates(inbox, &mut conn, &uid_store).await?;
-            Ok(())
+            let mut conn = main_conn.lock().await?;
+            conn.refresh(mailbox_hash).await
         }))
     }
 
@@ -456,6 +457,16 @@ impl MailBackend for ImapType {
             let new_mailboxes = Self::imap_mailboxes(&connection).await?;
             let mut mailboxes = uid_store.mailboxes.lock().await;
             *mailboxes = new_mailboxes;
+            for m in mailboxes.values_mut() {
+                log::trace!(
+                    "mailbox: {} is_subscribed: {}",
+                    m.path(),
+                    (uid_store.is_subscribed)(m.path())
+                );
+                if (uid_store.is_subscribed)(m.path()) {
+                    m.set_is_subscribed(true)?;
+                }
+            }
             /*
             let mut invalid_configs = vec![];
             for m in mailboxes.values() {
@@ -490,29 +501,19 @@ impl MailBackend for ImapType {
 
     fn is_online(&self) -> ResultFuture<()> {
         let connection = self.connection.clone();
-        let timeout_dur = self.server_conf.timeout;
         Ok(Box::pin(async move {
-            match timeout(timeout_dur, connection.lock()).await {
-                Ok(mut conn) => {
-                    imap_log!(trace, conn, "is_online");
-                    match timeout(timeout_dur, conn.connect()).await {
-                        Ok(Ok(())) => Ok(()),
-                        Err(err) | Ok(Err(err)) => {
-                            conn.stream = Err(err.clone());
-                            conn.connect().await
-                        }
-                    }
-                }
-                Err(err) => Err(err),
-            }
+            let mut conn = connection.lock().await?;
+            conn.is_online().await
         }))
     }
 
-    fn watch(&self) -> ResultFuture<()> {
+    fn watch(&self) -> ResultStream<BackendEvent> {
         let server_conf = self.server_conf.clone();
         let main_conn = self.connection.clone();
         let uid_store = self.uid_store.clone();
-        Ok(Box::pin(async move {
+        Ok(Box::pin(try_fn_stream(|emitter| async move {
+            use futures::stream::StreamExt;
+
             let has_idle: bool = match server_conf.protocol {
                 ImapProtocol::IMAP {
                     extension_use: ImapExtensionUse { idle, .. },
@@ -526,73 +527,101 @@ impl MailBackend for ImapType {
                 }
                 _ => false,
             };
-            while let Err(err) = if has_idle {
-                idle(ImapWatchKit {
+            enum WatchKit {
+                Idle(BoxStream<'static, Result<BackendEvent>>),
+                Poll(BoxStream<'static, Result<BackendEvent>>),
+            }
+            let mut watch_kit = if has_idle {
+                WatchKit::Idle(Box::pin(idle(ImapWatchKit {
                     conn: ImapConnection::new_connection(
                         &server_conf,
-                        "watch()::idle".into(),
+                        format!(
+                            "{}-watch-IDLE",
+                            uid_store.account_name.as_ref().trim_at_boundary(25)
+                        )
+                        .into(),
                         uid_store.clone(),
+                        false,
                     ),
                     main_conn: main_conn.clone(),
                     uid_store: uid_store.clone(),
-                })
-                .await
+                })))
             } else {
-                poll_with_examine(ImapWatchKit {
+                WatchKit::Poll(Box::pin(poll_with_examine(ImapWatchKit {
                     conn: ImapConnection::new_connection(
                         &server_conf,
-                        "watch()::poll_with_examine".into(),
+                        format!(
+                            "{}-watch-poll_with_EXAMINE",
+                            uid_store.account_name.as_ref().trim_at_boundary(25)
+                        )
+                        .into(),
                         uid_store.clone(),
+                        false,
                     ),
                     main_conn: main_conn.clone(),
                     uid_store: uid_store.clone(),
-                })
-                .await
+                })))
+            };
+            while let Some(ev) = {
+                match watch_kit {
+                    WatchKit::Idle(ref mut idle) => idle.next().await,
+                    WatchKit::Poll(ref mut poll) => poll.next().await,
+                }
             } {
-                let mut main_conn_lck = timeout(uid_store.timeout, main_conn.lock()).await?;
-                if err.kind.is_network() {
-                    uid_store.is_online.lock().unwrap().1 = Err(err.clone());
-                } else {
-                    return Err(err);
-                }
-                log::trace!(
-                    "{} Watch failure: {}",
-                    uid_store.account_name,
-                    err.to_string()
-                );
-                match timeout(uid_store.timeout, main_conn_lck.connect())
-                    .await
-                    .and_then(|res| res)
-                {
-                    Err(err2) => {
+                match ev {
+                    Ok(ok) => emitter.emit(ok).await,
+                    Err(err) => {
+                        let mut main_conn_lck = main_conn.lock().await?;
+                        if err.kind.is_network() {
+                            uid_store.is_online.lock().unwrap().1 = Err(err.clone());
+                        } else {
+                            return Err(err);
+                        }
                         log::trace!(
-                            "{} Watch reconnect attempt failed: {}",
+                            "{} Watch failure: {}",
                             uid_store.account_name,
-                            err2.to_string()
+                            err.to_string()
                         );
-                    }
-                    Ok(()) => {
-                        log::trace!(
-                            "{} Watch reconnect attempt successful",
-                            uid_store.account_name
-                        );
-                        continue;
+                        match timeout(uid_store.timeout, main_conn_lck.connect())
+                            .await
+                            .and_then(|res| res)
+                        {
+                            Err(err2) => {
+                                log::trace!(
+                                    "{} Watch reconnect attempt failed: {}",
+                                    uid_store.account_name,
+                                    err2.to_string()
+                                );
+                            }
+                            Ok(()) => {
+                                log::trace!(
+                                    "{} Watch reconnect attempt successful",
+                                    uid_store.account_name
+                                );
+                                continue;
+                            }
+                        }
+                        let account_hash = uid_store.account_hash;
+                        emitter
+                            .emit(
+                                RefreshEvent {
+                                    account_hash,
+                                    mailbox_hash: MailboxHash::default(),
+                                    kind: RefreshEventKind::Failure(err.clone()),
+                                }
+                                .into(),
+                            )
+                            .await;
+                        return Err(err);
                     }
                 }
-                let account_hash = uid_store.account_hash;
-                main_conn_lck.add_refresh_event(RefreshEvent {
-                    account_hash,
-                    mailbox_hash: MailboxHash::default(),
-                    kind: RefreshEventKind::Failure(err.clone()),
-                });
-                return Err(err);
             }
             log::trace!("{} watch future returning", uid_store.account_name);
             Ok(())
-        }))
+        })))
     }
 
-    fn operation(&self, hash: EnvelopeHash) -> Result<Box<dyn BackendOp>> {
+    fn envelope_bytes_by_hash(&self, hash: EnvelopeHash) -> ResultFuture<Vec<u8>> {
         let (uid, mailbox_hash) =
             if let Some(v) = self.uid_store.hash_index.lock().unwrap().get(&hash) {
                 *v
@@ -602,12 +631,13 @@ impl MailBackend for ImapType {
                      requested it.",
                 ));
             };
-        Ok(Box::new(ImapOp::new(
+        let op = ImapOp::new(
             uid,
             mailbox_hash,
             self.connection.clone(),
             self.uid_store.clone(),
-        )))
+        );
+        Ok(Box::pin(async move { op.as_bytes().await }))
     }
 
     fn save(
@@ -616,51 +646,10 @@ impl MailBackend for ImapType {
         mailbox_hash: MailboxHash,
         flags: Option<Flag>,
     ) -> ResultFuture<()> {
-        let uid_store = self.uid_store.clone();
         let connection = self.connection.clone();
         Ok(Box::pin(async move {
-            let mut response = Vec::with_capacity(8 * 1024);
-            let mut conn = connection.lock().await;
-            conn.select_mailbox(mailbox_hash, &mut response, true)
-                .await?;
-            let path = {
-                let mailboxes = uid_store.mailboxes.lock().await;
-
-                let mailbox = mailboxes.get(&mailbox_hash).ok_or_else(|| {
-                    Error::new(format!("Mailbox with hash {} not found.", mailbox_hash))
-                })?;
-                if !mailbox.permissions.lock().unwrap().create_messages {
-                    return Err(Error::new(format!(
-                        "You are not allowed to create messages in mailbox {}",
-                        mailbox.path()
-                    )));
-                }
-
-                mailbox.imap_path().to_string()
-            };
-            let flags = flags.unwrap_or_else(Flag::empty);
-            let has_literal_plus: bool = uid_store
-                .capabilities
-                .lock()
-                .unwrap()
-                .iter()
-                .any(|cap| cap.eq_ignore_ascii_case(b"LITERAL+"));
-            let data = if has_literal_plus {
-                Literal::try_from(bytes)?.into_non_sync()
-            } else {
-                Literal::try_from(bytes)?
-            };
-            conn.send_command(CommandBody::append(
-                path,
-                flags.derive_imap_codec_flags(),
-                None,
-                data,
-            )?)
-            .await?;
-            // [ref:TODO]: check for APPENDUID [RFC4315 - UIDPLUS]
-            conn.read_response(&mut response, RequiredResponses::empty())
-                .await?;
-            Ok(())
+            let mut conn = connection.lock().await?;
+            conn.save(bytes, mailbox_hash, flags).await
         }))
     }
 
@@ -671,38 +660,36 @@ impl MailBackend for ImapType {
         destination_mailbox_hash: MailboxHash,
         move_: bool,
     ) -> ResultFuture<()> {
-        let uid_store = self.uid_store.clone();
         let connection = self.connection.clone();
         let has_move: bool = move_
-            && uid_store
+            && self
+                .uid_store
                 .capabilities
                 .lock()
                 .unwrap()
                 .iter()
                 .any(|cap| cap.eq_ignore_ascii_case(b"MOVE"));
-        Ok(Box::pin(async move {
-            let uids: SmallVec<[UID; 64]> = {
-                let hash_index_lck = uid_store.hash_index.lock().unwrap();
-                env_hashes
-                    .iter()
-                    .filter_map(|env_hash| {
-                        hash_index_lck.get(&env_hash).cloned().map(|(uid, _)| uid)
-                    })
-                    .collect()
-            };
+        let uids: SmallVec<[UID; 64]> = {
+            let hash_index_lck = self.uid_store.hash_index.lock().unwrap();
+            env_hashes
+                .iter()
+                .filter_map(|env_hash| hash_index_lck.get(&env_hash).cloned().map(|(uid, _)| uid))
+                .collect()
+        };
 
-            if uids.is_empty() {
-                return Ok(());
-            }
+        if uids.is_empty() {
+            return Ok(Box::pin(async move { Ok(()) }));
+        }
+        Ok(Box::pin(async move {
+            let mut conn = connection.lock().await?;
             let dest_path = {
-                let mailboxes = uid_store.mailboxes.lock().await;
+                let mailboxes = conn.uid_store.mailboxes.lock().await;
                 let mailbox = mailboxes
                     .get(&destination_mailbox_hash)
                     .ok_or_else(|| Error::new("Destination mailbox not found"))?;
                 mailbox.imap_path().to_string()
             };
             let mut response = Vec::with_capacity(8 * 1024);
-            let mut conn = connection.lock().await;
             conn.select_mailbox(source_mailbox_hash, &mut response, false)
                 .await?;
             if has_move {
@@ -740,53 +727,50 @@ impl MailBackend for ImapType {
         flags: SmallVec<[FlagOp; 8]>,
     ) -> ResultFuture<()> {
         let connection = self.connection.clone();
-        let mut uid_store = self.uid_store.clone();
-        Ok(Box::pin(async move {
-            let uids: SmallVec<[UID; 64]> = {
-                let hash_index_lck = uid_store.hash_index.lock().unwrap();
-                env_hashes
-                    .iter()
-                    .filter_map(|env_hash| {
-                        hash_index_lck.get(&env_hash).cloned().map(|(uid, _)| uid)
-                    })
-                    .collect()
-            };
+        let uids: SmallVec<[UID; 64]> = {
+            let hash_index_lck = self.uid_store.hash_index.lock().unwrap();
+            env_hashes
+                .iter()
+                .filter_map(|env_hash| hash_index_lck.get(&env_hash).cloned().map(|(uid, _)| uid))
+                .collect()
+        };
 
-            if uids.is_empty() {
-                return Ok(());
+        if uids.is_empty() {
+            return Ok(Box::pin(async move { Ok(()) }));
+        }
+
+        // Conversion from UID (usize) to UID (NonZeroU32).
+        // TODO: Directly use NonZeroU32 for UID everywhere?
+        let sequence_set = {
+            let mut tmp = Vec::new();
+
+            for uid in uids.iter() {
+                if let Ok(uid) = u32::try_from(*uid) {
+                    if let Ok(uid) = NonZeroU32::try_from(uid) {
+                        tmp.push(uid);
+                    } else {
+                        return Err(Error::new("Application error: UID was 0"));
+                    }
+                } else {
+                    return Err(Error::new("Application error: UID exceeded u32"));
+                }
             }
 
-            // Conversion from UID (usize) to UID (NonZeroU32).
-            // TODO: Directly use NonZeroU32 for UID everywhere?
-            let sequence_set = {
-                let mut tmp = Vec::new();
-
-                for uid in uids.iter() {
-                    if let Ok(uid) = u32::try_from(*uid) {
-                        if let Ok(uid) = NonZeroU32::try_from(uid) {
-                            tmp.push(uid);
-                        } else {
-                            return Err(Error::new("Application error: UID was 0"));
-                        }
-                    } else {
-                        return Err(Error::new("Application error: UID exceeded u32"));
-                    }
-                }
-
-                // Safety: We know this can't fail due to the `is_empty` check above.
-                // TODO: This could be too easy to overlook.
-                SequenceSet::try_from(tmp).unwrap()
-            };
+            // Safety: We know this can't fail due to the `is_empty` check above.
+            // TODO: This could be too easy to overlook.
+            SequenceSet::try_from(tmp).unwrap()
+        };
 
+        Ok(Box::pin(async move {
             let mut response = Vec::with_capacity(8 * 1024);
-            let mut conn = connection.lock().await;
+            let mut conn = connection.lock().await?;
             conn.select_mailbox(mailbox_hash, &mut response, false)
                 .await?;
             if flags.iter().any(<bool>::from) {
                 /* Set flags/tags to true */
                 let mut set_seen = false;
                 let command = {
-                    let mut tag_lck = uid_store.collection.tag_index.write().unwrap();
+                    let mut tag_lck = conn.uid_store.collection.tag_index.write().unwrap();
                     let flags = {
                         let mut tmp = Vec::new();
 
@@ -845,7 +829,7 @@ impl MailBackend for ImapType {
                 conn.read_response(&mut response, RequiredResponses::empty())
                     .await?;
                 if set_seen {
-                    for f in uid_store.mailboxes.lock().await.values() {
+                    for f in conn.uid_store.mailboxes.lock().await.values() {
                         if let Ok(mut unseen) = f.unseen.lock() {
                             for env_hash in env_hashes.iter() {
                                 unseen.remove(env_hash);
@@ -913,7 +897,7 @@ impl MailBackend for ImapType {
                 conn.read_response(&mut response, RequiredResponses::empty())
                     .await?;
                 if set_unseen {
-                    for f in uid_store.mailboxes.lock().await.values() {
+                    for f in conn.uid_store.mailboxes.lock().await.values() {
                         if let (Ok(mut unseen), Ok(exists)) = (f.unseen.lock(), f.exists.lock()) {
                             for env_hash in env_hashes.iter().filter(|h| exists.contains(h)) {
                                 unseen.insert_new(env_hash);
@@ -922,7 +906,8 @@ impl MailBackend for ImapType {
                     }
                 }
             }
-            if let Err(err) = uid_store
+            if let Err(err) = conn
+                .uid_store
                 .update_flags(env_hashes, mailbox_hash, flags)
                 .or_else(sync::cache::ignore_not_found)
             {
@@ -947,7 +932,7 @@ impl MailBackend for ImapType {
         Ok(Box::pin(async move {
             flag_future.await?;
             let mut response = Vec::with_capacity(8 * 1024);
-            let mut conn = connection.lock().await;
+            let mut conn = connection.lock().await?;
             let has_uid_plus = conn
                 .uid_store
                 .capabilities
@@ -1079,7 +1064,7 @@ impl MailBackend for ImapType {
 
             let mut response = Vec::with_capacity(8 * 1024);
             {
-                let mut conn_lck = connection.lock().await;
+                let mut conn_lck = connection.lock().await?;
                 conn_lck.unselect().await?;
 
                 conn_lck
@@ -1139,7 +1124,7 @@ impl MailBackend for ImapType {
             }
             let mut response = Vec::with_capacity(8 * 1024);
             {
-                let mut conn_lck = connection.lock().await;
+                let mut conn_lck = connection.lock().await?;
                 /* make sure mailbox is not selected before it gets deleted, otherwise
                  * connection gets dropped by server */
                 conn_lck.unselect().await?;
@@ -1192,7 +1177,7 @@ impl MailBackend for ImapType {
 
             let mut response = Vec::with_capacity(8 * 1024);
             {
-                let mut conn_lck = connection.lock().await;
+                let mut conn_lck = connection.lock().await?;
                 if new_val {
                     conn_lck
                         .send_command(CommandBody::subscribe(imap_path.as_str())?)
@@ -1256,7 +1241,7 @@ impl MailBackend for ImapType {
                 command = CommandBody::Rename { from, to };
             }
             {
-                let mut conn_lck = connection.lock().await;
+                let mut conn_lck = connection.lock().await?;
                 conn_lck.send_command(command).await?;
                 conn_lck
                     .read_response(&mut response, RequiredResponses::empty())
@@ -1305,7 +1290,7 @@ impl MailBackend for ImapType {
 
     fn search(
         &self,
-        query: crate::search::Query,
+        query: Query,
         mailbox_hash: Option<MailboxHash>,
     ) -> ResultFuture<Vec<EnvelopeHash>> {
         if mailbox_hash.is_none() {
@@ -1313,44 +1298,11 @@ impl MailBackend for ImapType {
                 "Cannot search without specifying mailbox on IMAP",
             ));
         }
-        let mailbox_hash = mailbox_hash.unwrap();
-        let query_str = query.to_imap_search();
         let connection = self.connection.clone();
-        let uid_store = self.uid_store.clone();
 
         Ok(Box::pin(async move {
-            let mut response = Vec::with_capacity(8 * 1024);
-            let mut conn = connection.lock().await;
-            conn.examine_mailbox(mailbox_hash, &mut response, false)
-                .await?;
-            conn.send_command_raw(
-                format!("UID SEARCH CHARSET UTF-8 {}", query_str.trim()).as_bytes(),
-            )
-            .await?;
-            conn.read_response(&mut response, RequiredResponses::SEARCH)
-                .await?;
-            imap_log!(
-                trace,
-                conn,
-                "searching for {} returned: {}",
-                query_str,
-                String::from_utf8_lossy(&response)
-            );
-
-            for l in response.split_rn() {
-                if l.starts_with(b"* SEARCH") {
-                    let uid_index = uid_store.uid_index.lock()?;
-                    return Ok(Vec::from_iter(
-                        String::from_utf8_lossy(l[b"* SEARCH".len()..].trim())
-                            .split_whitespace()
-                            .map(UID::from_str)
-                            .filter_map(std::result::Result::ok)
-                            .filter_map(|uid| uid_index.get(&(mailbox_hash, uid)))
-                            .copied(),
-                    ));
-                }
-            }
-            Err(Error::new(String::from_utf8_lossy(&response).to_string()))
+            let mut conn = connection.lock().await?;
+            conn.search(query, mailbox_hash).await
         }))
     }
 }
@@ -1401,6 +1353,7 @@ impl ImapType {
         } else {
             Some(Duration::from_secs(timeout))
         };
+        let use_connection_pool = get_conf_val!(s["use_connection_pool"], true)?;
         let server_conf = ImapServerConf {
             server_hostname: server_hostname.to_string(),
             server_username: server_username.to_string(),
@@ -1426,6 +1379,7 @@ impl ImapType {
         let uid_store: Arc<UIDStore> = Arc::new(UIDStore {
             offline_cache: Arc::new(Mutex::new(None)),
             ..UIDStore::new(
+                is_subscribed,
                 account_hash,
                 account_name,
                 event_consumer,
@@ -1433,13 +1387,25 @@ impl ImapType {
                 keep_offline_cache,
             )
         });
-        let connection =
-            ImapConnection::new_connection(&server_conf, "ImapType::new".into(), uid_store.clone());
+        let connection = ImapConnection::new_connection(
+            &server_conf,
+            format!(
+                "{}-main-conn",
+                uid_store.account_name.as_ref().trim_at_boundary(25),
+            )
+            .into(),
+            uid_store.clone(),
+            true,
+        );
 
         Ok(Box::new(Self {
+            connection: Arc::new(ConnectionMutex::new(
+                connection,
+                server_conf.clone(),
+                uid_store.clone(),
+                use_connection_pool,
+            )),
             server_conf,
-            _is_subscribed: is_subscribed,
-            connection: Arc::new(FutureMutex::new(connection)),
             uid_store,
         }))
     }
@@ -1447,8 +1413,13 @@ impl ImapType {
     pub fn shell(&self) {
         let mut conn = ImapConnection::new_connection(
             &self.server_conf,
-            "ImapType::shell".into(),
+            format!(
+                "{}-shell-conn",
+                self.uid_store.account_name.as_ref().trim_at_boundary(25),
+            )
+            .into(),
             self.uid_store.clone(),
+            true,
         );
 
         futures::executor::block_on(timeout(self.server_conf.timeout, conn.connect()))
@@ -1507,11 +1478,11 @@ impl ImapType {
     }
 
     pub async fn imap_mailboxes(
-        connection: &Arc<FutureMutex<ImapConnection>>,
+        connection: &Arc<ConnectionMutex>,
     ) -> Result<HashMap<MailboxHash, ImapMailbox>> {
         let mut mailboxes: HashMap<MailboxHash, ImapMailbox> = Default::default();
         let mut res = Vec::with_capacity(8 * 1024);
-        let mut conn = connection.lock().await;
+        let mut conn = connection.lock().await?;
         let has_list_status: bool = conn
             .uid_store
             .capabilities
@@ -1525,12 +1496,12 @@ impl ImapType {
                 .await?;
             conn.read_response(
                 &mut res,
-                RequiredResponses::LIST_REQUIRED | RequiredResponses::STATUS,
+                RequiredResponses::LIST | RequiredResponses::STATUS,
             )
             .await?;
         } else {
             conn.send_command(CommandBody::list("", "*")?).await?;
-            conn.read_response(&mut res, RequiredResponses::LIST_REQUIRED)
+            conn.read_response(&mut res, RequiredResponses::LIST)
                 .await?;
         }
         imap_log!(trace, conn, "LIST reply: {}", String::from_utf8_lossy(&res));
@@ -1580,7 +1551,7 @@ impl ImapType {
         }
         mailboxes.retain(|_, v| !v.hash.is_null());
         conn.send_command(CommandBody::lsub("", "*")?).await?;
-        conn.read_response(&mut res, RequiredResponses::LSUB_REQUIRED)
+        conn.read_response(&mut res, RequiredResponses::LSUB)
             .await?;
         imap_log!(trace, conn, "LSUB reply: {}", String::from_utf8_lossy(&res));
         for l in res.split_rn() {
@@ -1690,6 +1661,7 @@ impl ImapType {
         get_conf_val!(s["use_auth_anonymous"], false)?;
         get_conf_val!(s["use_id"], false)?;
         let _timeout = get_conf_val!(s["timeout"], 16_u64)?;
+        get_conf_val!(s["use_connection_pool"], true)?;
         let extra_keys = s
             .extra
             .keys()
