@@ -24,36 +24,30 @@ use std::convert::{TryFrom, TryInto};
 use imap_codec::imap_types::{command::CommandBody, search::SearchKey, sequence::SequenceSet};
 
 use crate::{
-    backends::{
-        BackendMailbox, RefreshEvent,
-        RefreshEventKind::{self, *},
-        TagHash,
-    },
-    email::common_attributes,
+    backends::{BackendEvent, BackendMailbox, RefreshEvent, RefreshEventKind::*, TagHash},
     error::*,
     imap::{
+        email::common_attributes,
         protocol_parser::{
             generate_envelope_hash, FetchResponse, ImapLineSplit, RequiredResponses,
             UntaggedResponse,
         },
-        sync::cache::ignore_not_found,
+        sync::cache::{ignore_not_found, ImapCache},
         ImapConnection, MailboxSelection, UID,
     },
 };
 
 impl ImapConnection {
-    pub async fn process_untagged(&mut self, line: &[u8]) -> Result<bool> {
+    pub async fn process_untagged(
+        &mut self,
+        untagged_response: UntaggedResponse<'_>,
+    ) -> Result<Option<BackendEvent>> {
         macro_rules! try_fail {
             ($mailbox_hash: expr, $($result:expr $(,)*)+) => {
                 $(if let Err(err) = $result {
                     self.uid_store.is_online.lock().unwrap().1 = Err(err.clone());
                     imap_log!(trace, self, "failure: {}", err.to_string());
                     log::debug!("failure: {}", err.to_string());
-                    self.add_refresh_event(RefreshEvent {
-                        account_hash: self.uid_store.account_hash,
-                        mailbox_hash: $mailbox_hash,
-                        kind: RefreshEventKind::Failure(err.clone()),
-                    });
                     Err(err)
                 } else { Ok(()) }?;)+
             };
@@ -65,22 +59,16 @@ impl ImapConnection {
             | MailboxSelection::Examine {
                 mailbox_hash: h, ..
             } => h,
-            MailboxSelection::None => return Ok(false),
+            MailboxSelection::None => return Ok(None),
         };
         let mailbox =
             std::clone::Clone::clone(&self.uid_store.mailboxes.lock().await[&mailbox_hash]);
 
         let mut response = Vec::with_capacity(8 * 1024);
-        let untagged_response =
-            match super::protocol_parser::untagged_responses(line).map(|(_, v, _)| v) {
-                Ok(None) | Err(_) => {
-                    return Ok(false);
-                }
-                Ok(Some(r)) => r,
-            };
         match untagged_response {
             UntaggedResponse::Bye { reason } => {
                 self.uid_store.is_online.lock().unwrap().1 = Err(reason.into());
+                Err(reason.into())
             }
             UntaggedResponse::Expunge(n) => {
                 if self
@@ -120,61 +108,64 @@ impl ImapConnection {
                                 .enumerate(),
                         );
                     }
-                    let mut events = vec![];
-                    let deleteds = self
-                        .uid_store
-                        .uid_index
-                        .lock()
-                        .unwrap()
-                        .iter()
-                        .filter(|((_, u), _)| !results.contains(u))
-                        .map(|((_, uid), hash)| (*uid, *hash))
-                        .collect::<Vec<(UID, crate::email::EnvelopeHash)>>();
-                    let mut mboxes_to_update = vec![];
-                    for (deleted_uid, deleted_hash) in deleteds {
-                        for (mailbox_hash, mbx) in self.uid_store.mailboxes.lock().await.iter_mut()
-                        {
-                            mbx.exists.lock().unwrap().remove(deleted_hash);
-                            mbx.unseen.lock().unwrap().remove(deleted_hash);
-                            let mut existed = self
-                                .uid_store
+                    let mut uid_events = vec![];
+                    {
+                        let mut mboxes = self.uid_store.mailboxes.lock().await;
+                        let deleted_uids_hashes = self
+                            .uid_store
+                            .uid_index
+                            .lock()
+                            .unwrap()
+                            .iter()
+                            .filter(|((mbx, u), _)| *mbx == mailbox_hash && !results.contains(u))
+                            .map(|((_, uid), hash)| (*uid, *hash))
+                            .collect::<Vec<(UID, crate::email::EnvelopeHash)>>();
+                        for (deleted_uid, deleted_hash) in deleted_uids_hashes {
+                            self.uid_store
                                 .uid_index
                                 .lock()
                                 .unwrap()
-                                .remove(&(*mailbox_hash, deleted_uid))
-                                .is_some();
-                            existed |= self
-                                .uid_store
+                                .remove(&(mailbox_hash, deleted_uid));
+                            {
+                                if let Some(mbx) = mboxes.get_mut(&mailbox_hash) {
+                                    mbx.exists.lock().unwrap().remove(deleted_hash);
+                                    mbx.unseen.lock().unwrap().remove(deleted_hash);
+                                }
+                            }
+                            self.uid_store
                                 .hash_index
                                 .lock()
                                 .unwrap()
-                                .remove(&deleted_hash)
-                                .is_some();
-                            if existed {
-                                mboxes_to_update.push(*mailbox_hash);
-                                events.push((
-                                    deleted_uid,
-                                    RefreshEvent {
-                                        account_hash: self.uid_store.account_hash,
-                                        mailbox_hash: *mailbox_hash,
-                                        kind: Remove(deleted_hash),
-                                    },
-                                ));
-                            }
+                                .remove(&deleted_hash);
+                            uid_events.push((
+                                deleted_uid,
+                                RefreshEvent {
+                                    account_hash: self.uid_store.account_hash,
+                                    mailbox_hash,
+                                    kind: Remove(deleted_hash),
+                                },
+                            ));
                         }
                     }
-                    for mailbox_hash in mboxes_to_update {
-                        use crate::imap::sync::cache::ImapCache;
-
-                        self.uid_store
-                            .update(mailbox_hash, &events)
-                            .or_else(ignore_not_found)?;
+                    if let Err(err) = self
+                        .uid_store
+                        .update(mailbox_hash, &uid_events)
+                        .or_else(ignore_not_found)
+                    {
+                        log::error!(
+                            "Could not update cache for mailbox_hash = {:?} uid, events = {:?}: \
+                             err = {}",
+                            mailbox_hash,
+                            uid_events,
+                            err
+                        );
                     }
-                    for (_, event) in events {
-                        self.add_refresh_event(event);
-                    }
-
-                    return Ok(true);
+                    return Ok(uid_events
+                        .into_iter()
+                        .map(|(_, ev)| ev)
+                        .collect::<Vec<_>>()
+                        .try_into()
+                        .ok());
                 }
                 let Some(deleted_uid) = self
                     .uid_store
@@ -185,63 +176,63 @@ impl ImapConnection {
                     .or_default()
                     .remove(&TryInto::<usize>::try_into(n).unwrap().saturating_sub(1))
                 else {
-                    return Ok(true);
+                    return Ok(None);
                 };
                 imap_log!(trace, self, "expunge {}, UID = {}", n, deleted_uid);
-                let deleted_hash: crate::email::EnvelopeHash = match self
+                let Some(deleted_hash) = self
                     .uid_store
                     .uid_index
                     .lock()
                     .unwrap()
                     .remove(&(mailbox_hash, deleted_uid))
-                {
-                    Some(v) => v,
-                    None => return Ok(true),
+                else {
+                    return Ok(None);
                 };
-                let mut mboxes_to_update = vec![];
-                for (mailbox_hash, mbx) in self.uid_store.mailboxes.lock().await.iter_mut() {
-                    if mbx.exists.lock().unwrap().remove(deleted_hash) {
-                        mboxes_to_update.push(*mailbox_hash);
+                {
+                    let mut mboxes = self.uid_store.mailboxes.lock().await;
+                    if let Some(mbx) = mboxes.get_mut(&mailbox_hash) {
+                        mbx.exists.lock().unwrap().remove(deleted_hash);
+                        mbx.unseen.lock().unwrap().remove(deleted_hash);
                     }
-                    mbx.unseen.lock().unwrap().remove(deleted_hash);
                 }
                 self.uid_store
                     .hash_index
                     .lock()
                     .unwrap()
                     .remove(&deleted_hash);
-                let events = mboxes_to_update
-                    .into_iter()
-                    .map(|mailbox_hash| {
-                        (
-                            mailbox_hash,
-                            [(
-                                deleted_uid,
-                                RefreshEvent {
-                                    account_hash: self.uid_store.account_hash,
-                                    mailbox_hash,
-                                    kind: Remove(deleted_hash),
-                                },
-                            )],
-                        )
-                    })
-                    .collect::<Vec<(_, [(UID, RefreshEvent); 1])>>();
-                for (mailbox_hash, pair) in events {
-                    use crate::imap::sync::cache::ImapCache;
+                let pair = [(
+                    deleted_uid,
+                    RefreshEvent {
+                        account_hash: self.uid_store.account_hash,
+                        mailbox_hash,
+                        kind: Remove(deleted_hash),
+                    },
+                )];
 
-                    self.uid_store
-                        .update(mailbox_hash, &pair)
-                        .or_else(ignore_not_found)?;
-                    let [(_, event)] = pair;
-                    self.add_refresh_event(event);
+                if let Err(err) = self
+                    .uid_store
+                    .update(mailbox_hash, &pair)
+                    .or_else(ignore_not_found)
+                {
+                    log::error!(
+                        "Could not update cache for mailbox_hash = {:?} uid = {:?}, events = \
+                         {:?}: err = {}",
+                        mailbox_hash,
+                        deleted_uid,
+                        pair,
+                        err
+                    );
                 }
+                let [(_, event)] = pair;
+                Ok(Some(event.into()))
             }
             UntaggedResponse::Exists(n) => {
                 imap_log!(trace, self, "exists {}", n);
+                let (required_responses, attributes) = common_attributes();
                 try_fail!(
                     mailbox_hash,
-                    self.send_command(CommandBody::fetch(n, common_attributes(), false)?).await
-                    self.read_response(&mut response, RequiredResponses::FETCH_REQUIRED).await
+                    self.send_command(CommandBody::fetch(n, attributes, false)?).await
+                    self.read_response(&mut response, required_responses).await
                 );
                 let mut v = match super::protocol_parser::fetch_responses(&response) {
                     Ok((_, v, _)) => v,
@@ -252,7 +243,7 @@ impl ImapConnection {
                             "Error when parsing FETCH response after untagged exists {:?}",
                             err
                         );
-                        return Ok(true);
+                        return Ok(None);
                     }
                 };
                 imap_log!(trace, self, "responses len is {}", v.len());
@@ -300,7 +291,7 @@ impl ImapConnection {
                             .unwrap()
                             .entry(mailbox_hash)
                             .or_default()
-                            .insert(*message_sequence_number, uid);
+                            .insert(message_sequence_number.saturating_sub(1), uid);
                     }
                     self.uid_store
                         .hash_index
@@ -322,8 +313,6 @@ impl ImapConnection {
                     );
                 }
                 {
-                    use crate::imap::sync::cache::ImapCache;
-
                     if let Err(err) = self
                         .uid_store
                         .insert_envelopes(mailbox_hash, &v)
@@ -338,19 +327,21 @@ impl ImapConnection {
                         imap_log!(info, self, "{}", err);
                     }
                 }
+                let mut events = vec![];
                 for response in v {
                     if let FetchResponse {
                         envelope: Some(envelope),
                         ..
                     } = response
                     {
-                        self.add_refresh_event(RefreshEvent {
+                        events.push(RefreshEvent {
                             account_hash: self.uid_store.account_hash,
                             mailbox_hash,
                             kind: Create(Box::new(envelope)),
                         });
                     }
                 }
+                Ok(events.try_into().ok())
             }
             UntaggedResponse::Recent(_) => {
                 try_fail!(
@@ -364,8 +355,10 @@ impl ImapConnection {
                 {
                     Ok(&[]) => {
                         imap_log!(trace, self, "UID SEARCH RECENT returned no results");
+                        Ok(None)
                     }
                     Ok(v) => {
+                        // [ref:FIXME]: use imap_codec types instead of a raw command
                         let command = {
                             let mut iter = v.split(u8::is_ascii_whitespace);
                             let first = iter.next().unwrap_or(v);
@@ -382,8 +375,16 @@ impl ImapConnection {
                         };
                         try_fail!(
                             mailbox_hash,
-                            self.send_command_raw(command.as_bytes()).await
-                            self.read_response(&mut response, RequiredResponses::FETCH_REQUIRED).await
+                            self.send_command_raw(command.as_bytes()).await,
+                            self.read_response(
+                                &mut response,
+                                RequiredResponses::FETCH_UID
+                                    | RequiredResponses::FETCH_FLAGS
+                                    | RequiredResponses::FETCH_ENVELOPE
+                                    | RequiredResponses::FETCH_REFERENCES
+                                    | RequiredResponses::FETCH_BODYSTRUCTURE,
+                            )
+                            .await,
                         );
                         let mut v = match super::protocol_parser::fetch_responses(&response) {
                             Ok((_, v, _)) => v,
@@ -394,7 +395,7 @@ impl ImapConnection {
                                     "Error when parsing FETCH response after untagged recent {:?}",
                                     err
                                 );
-                                return Ok(true);
+                                return Ok(None);
                             }
                         };
                         imap_log!(trace, self, "responses len is {}", v.len());
@@ -430,8 +431,6 @@ impl ImapConnection {
                             mailbox.exists.lock().unwrap().insert_new(env.hash());
                         }
                         {
-                            use crate::imap::sync::cache::ImapCache;
-
                             if let Err(err) = self
                                 .uid_store
                                 .insert_envelopes(mailbox_hash, &v)
@@ -446,6 +445,7 @@ impl ImapConnection {
                                 log::info!("{err}");
                             }
                         }
+                        let mut events = vec![];
                         for response in v {
                             if let FetchResponse {
                                 envelope: Some(envelope),
@@ -467,7 +467,7 @@ impl ImapConnection {
                                         .unwrap()
                                         .entry(mailbox_hash)
                                         .or_default()
-                                        .insert(message_sequence_number, uid);
+                                        .insert(message_sequence_number.saturating_sub(1), uid);
                                 }
                                 self.uid_store
                                     .hash_index
@@ -485,20 +485,22 @@ impl ImapConnection {
                                     envelope.subject(),
                                     mailbox.path(),
                                 );
-                                self.add_refresh_event(RefreshEvent {
+                                events.push(RefreshEvent {
                                     account_hash: self.uid_store.account_hash,
                                     mailbox_hash,
                                     kind: Create(Box::new(envelope)),
                                 });
                             }
                         }
+                        Ok(events.try_into().ok())
                     }
-                    Err(e) => {
+                    Err(err) => {
                         debug!(
                             "UID SEARCH RECENT err: {}\nresp: {}",
-                            e.to_string(),
+                            err,
                             to_str!(&response)
                         );
+                        Ok(None)
                     }
                 }
             }
@@ -510,6 +512,7 @@ impl ImapConnection {
                 body: _,
                 references: _,
                 envelope: _,
+                bodystructure: _,
                 raw_fetch_value: _,
             }) => {
                 if let Some(flags) = flags {
@@ -534,12 +537,12 @@ impl ImapConnection {
                         {
                             Ok(mut v) if v.len() == 1 => v.pop().unwrap(),
                             Ok(_) => {
-                                return Ok(false);
+                                return Ok(None);
                             }
                             Err(e) => {
                                 debug!("SEARCH error failed: {}", e);
                                 debug!(to_str!(&response));
-                                return Ok(false);
+                                return Ok(None);
                             }
                         }
                     };
@@ -573,7 +576,7 @@ impl ImapConnection {
                                 .unwrap()
                                 .insert(env_hash, modseq);
                         }
-                        let mut event: [(UID, RefreshEvent); 1] = [(
+                        let event: [(UID, RefreshEvent); 1] = [(
                             uid,
                             RefreshEvent {
                                 account_hash: self.uid_store.account_hash,
@@ -582,22 +585,17 @@ impl ImapConnection {
                             },
                         )];
                         {
-                            use crate::imap::sync::cache::ImapCache;
-
                             self.uid_store.update(mailbox_hash, &event)?;
                         }
-                        self.add_refresh_event(std::mem::replace(
-                            &mut event[0].1,
-                            RefreshEvent {
-                                account_hash: self.uid_store.account_hash,
-                                mailbox_hash,
-                                kind: Rescan,
-                            },
-                        ));
-                    };
+                        let [(_, ev)] = event;
+                        Ok(Some(ev.into()))
+                    } else {
+                        Ok(None)
+                    }
+                } else {
+                    Ok(None)
                 }
             }
         }
-        Ok(true)
     }
 }
