@@ -45,7 +45,7 @@ use crate::{
             EmailQueryChangesResponse,
         },
         filters::Filter,
-        identity::{IdentityGet, IdentityObject, IdentitySet},
+        identity::{Identity, IdentityGet, IdentitySet},
         mailbox::MailboxObject,
         methods::{
             AddedItem, Changes, ChangesResponse, Get, GetResponse, MethodResponse, QueryChanges,
@@ -314,6 +314,10 @@ impl JmapConnection {
             .unwrap()
             .clone_from(&session.capabilities);
         let mail_account_id = session.mail_account_id();
+        {
+            let mut metadata = self.store.metadata.lock().unwrap();
+            metadata.insert("session".into(), serde_json::json! {session});
+        };
         _ = self
             .store
             .online_status
@@ -342,14 +346,14 @@ impl JmapConnection {
                 }
                 Ok(s) => s,
             };
-            let GetResponse::<IdentityObject> { list, .. } =
-                GetResponse::<IdentityObject>::try_from(v.method_responses.remove(0))?;
+            let GetResponse::<Identity> { list, .. } =
+                GetResponse::<Identity>::try_from(v.method_responses.remove(0))?;
             list
         };
         if id_list.is_empty() {
             let mut req = Request::new(self.request_no.clone());
             let identity_set = IdentitySet(
-                Set::<IdentityObject>::new(None)
+                Set::<Identity>::new(None)
                     .account_id(mail_account_id.clone())
                     .create(Some({
                         let address =
@@ -360,23 +364,23 @@ impl JmapConnection {
                                         self.store.main_identity.clone(),
                                     )
                                 });
-                        let id: Id<IdentityObject> = Id::new_uuid_v4();
+                        let id: Id<Identity> = Id::new_random();
                         log::trace!(
                             "identity id = {}, {:#?}",
                             id,
-                            IdentityObject {
+                            Identity {
                                 id: id.clone(),
                                 name: address.get_display_name().unwrap_or_default(),
                                 email: address.get_email(),
-                                ..IdentityObject::default()
+                                ..Identity::default()
                             }
                         );
                         indexmap! {
-                            id.clone().into() => IdentityObject {
+                            id.clone().into() => Identity {
                                 id,
                                 name: address.get_display_name().unwrap_or_default(),
                                 email: address.get_email(),
-                                ..IdentityObject::default()
+                                ..Identity::default()
                             }
                         }
                     })),
@@ -417,8 +421,8 @@ impl JmapConnection {
                 }
                 Ok(s) => s,
             };
-            let GetResponse::<IdentityObject> { list, .. } =
-                GetResponse::<IdentityObject>::try_from(v.method_responses.remove(0))?;
+            let GetResponse::<Identity> { list, .. } =
+                GetResponse::<Identity>::try_from(v.method_responses.remove(0))?;
             id_list = list;
         }
         self.session_guard().await?.identities =
@@ -435,22 +439,19 @@ impl JmapConnection {
     }
 
     #[inline]
-    pub fn add_refresh_event(&self, ev: RefreshEvent) {
-        self.add_backend_event(BackendEvent::Refresh(ev));
-    }
-
     pub fn add_backend_event(&self, ev: BackendEvent) {
         (self.store.event_consumer)(self.store.account_hash, ev);
     }
 
-    pub async fn email_changes(&self, mailbox_hash: MailboxHash) -> Result<()> {
+    pub async fn email_changes(&self, mailbox_hash: MailboxHash) -> Result<Option<BackendEvent>> {
         let mut current_state: State<EmailObject> =
             if let Some(s) = self.store.email_state.lock().await.clone() {
                 s
             } else {
-                return Ok(());
+                return Ok(None);
             };
         let mail_account_id = self.session_guard().await?.mail_account_id();
+        let mut events = vec![];
         loop {
             let email_changes_call: EmailChanges = EmailChanges::new(
                 Changes::<EmailObject>::new()
@@ -489,9 +490,7 @@ impl JmapConnection {
                 let email_query_changes_call = EmailQueryChanges::new(
                     QueryChanges::new(mail_account_id.clone(), email_query_state).filter(Some(
                         Filter::Condition(
-                            EmailFilterCondition::new()
-                                .in_mailbox(Some(mailbox_id.clone()))
-                                .into(),
+                            EmailFilterCondition::new().in_mailbox(Some(mailbox_id.clone())),
                         ),
                     )),
                 );
@@ -510,8 +509,6 @@ impl JmapConnection {
                         .properties(Some(vec!["keywords".to_string(), "mailboxIds".to_string()])),
                 );
                 req.add_call(&email_get_call).await;
-            } else {
-                return Ok(());
             }
             let res_text = self
                 .post_async(None, serde_json::to_string(&req)?)
@@ -531,15 +528,28 @@ impl JmapConnection {
                 }
                 Ok(s) => s,
             };
-            let changes_response =
+            let mut changes_response =
                 ChangesResponse::<EmailObject>::try_from(v.method_responses.remove(0))?;
             if changes_response.new_state == current_state {
-                return Ok(());
+                return Ok(None);
             }
+            for destroyed_id in std::mem::take(&mut changes_response.destroyed) {
+                if let Some((env_hash, mailbox_hashes)) =
+                    self.store.remove_envelope(destroyed_id).await
+                {
+                    for mailbox_hash in mailbox_hashes {
+                        events.push(RefreshEvent {
+                            account_hash: self.store.account_hash,
+                            mailbox_hash,
+                            kind: RefreshEventKind::Remove(env_hash),
+                        });
+                    }
+                }
+            }
+            // [ref:TODO]: process changes_response.updated too
             let get_response = GetResponse::<EmailObject>::try_from(v.method_responses.remove(0))?;
 
             {
-                /* process get response */
                 let GetResponse::<EmailObject> { list, .. } = get_response;
 
                 let mut mailbox_hashes: Vec<SmallVec<[MailboxHash; 8]>> =
@@ -566,7 +576,7 @@ impl JmapConnection {
                             }
                             mbox.total_emails.lock().unwrap().insert_new(env.hash());
                         });
-                        self.add_refresh_event(RefreshEvent {
+                        events.push(RefreshEvent {
                             account_hash: self.store.account_hash,
                             mailbox_hash,
                             kind: RefreshEventKind::Create(Box::new(env.clone())),
@@ -580,7 +590,7 @@ impl JmapConnection {
                             }
                             mbox.total_emails.lock().unwrap().insert_new(env.hash());
                         });
-                        self.add_refresh_event(RefreshEvent {
+                        events.push(RefreshEvent {
                             account_hash: self.store.account_hash,
                             mailbox_hash,
                             kind: RefreshEventKind::Create(Box::new(env)),
@@ -589,93 +599,95 @@ impl JmapConnection {
                 }
             }
             let reverse_id_store_lck = self.store.reverse_id_store.lock().await;
-            let response = v.method_responses.remove(0);
-            match EmailQueryChangesResponse::try_from(response) {
-                Ok(EmailQueryChangesResponse {
-                    collapse_threads: _,
-                    query_changes_response:
-                        QueryChangesResponse {
-                            account_id: _,
-                            old_query_state,
-                            new_query_state,
-                            total: _,
-                            removed,
-                            added,
-                        },
-                }) if old_query_state != new_query_state => {
-                    self.store
-                        .mailboxes
-                        .write()
-                        .unwrap()
-                        .entry(mailbox_hash)
-                        .and_modify(|mbox| {
-                            *mbox.email_query_state.lock().unwrap() = Some(new_query_state);
-                        });
-                    /*  If the "filter" or "sort" includes a mutable property, the server
-                    MUST include all Foos in the current results for which this
-                    property may have changed.  The position of these may have moved
-                    in the results, so they must be reinserted by the client to ensure
-                    its query cache is correct.  */
-                    for email_obj_id in removed
-                        .into_iter()
-                        .filter(|id| !added.iter().any(|item| item.id == *id))
-                    {
-                        if let Some(env_hash) = reverse_id_store_lck.get(&email_obj_id) {
-                            let mut mailboxes_lck = self.store.mailboxes.write().unwrap();
-                            mailboxes_lck.entry(mailbox_hash).and_modify(|mbox| {
-                                mbox.unread_emails.lock().unwrap().remove(*env_hash);
-                                mbox.total_emails.lock().unwrap().insert_new(*env_hash);
-                            });
-                            self.add_refresh_event(RefreshEvent {
-                                account_hash: self.store.account_hash,
-                                mailbox_hash,
-                                kind: RefreshEventKind::Remove(*env_hash),
+            if !v.method_responses.is_empty() {
+                let response = v.method_responses.remove(0);
+                match EmailQueryChangesResponse::try_from(response) {
+                    Ok(EmailQueryChangesResponse {
+                        collapse_threads: _,
+                        query_changes_response:
+                            QueryChangesResponse {
+                                account_id: _,
+                                old_query_state,
+                                new_query_state,
+                                total: _,
+                                removed,
+                                added,
+                            },
+                    }) if old_query_state != new_query_state => {
+                        self.store
+                            .mailboxes
+                            .write()
+                            .unwrap()
+                            .entry(mailbox_hash)
+                            .and_modify(|mbox| {
+                                *mbox.email_query_state.lock().unwrap() = Some(new_query_state);
                             });
+                        /*  If the "filter" or "sort" includes a mutable property, the server
+                        MUST include all Foos in the current results for which this
+                        property may have changed.  The position of these may have moved
+                        in the results, so they must be reinserted by the client to ensure
+                        its query cache is correct.  */
+                        for email_obj_id in removed
+                            .into_iter()
+                            .filter(|id| !added.iter().any(|item| item.id == *id))
+                        {
+                            if let Some(env_hash) = reverse_id_store_lck.get(&email_obj_id) {
+                                let mut mailboxes_lck = self.store.mailboxes.write().unwrap();
+                                mailboxes_lck.entry(mailbox_hash).and_modify(|mbox| {
+                                    mbox.unread_emails.lock().unwrap().remove(*env_hash);
+                                    mbox.total_emails.lock().unwrap().insert_new(*env_hash);
+                                });
+                                events.push(RefreshEvent {
+                                    account_hash: self.store.account_hash,
+                                    mailbox_hash,
+                                    kind: RefreshEventKind::Remove(*env_hash),
+                                });
+                            }
+                        }
+                        for AddedItem {
+                            id: _email_obj_id,
+                            index: _,
+                        } in added
+                        {
+                            // [ref:TODO] do something with added items
                         }
                     }
-                    for AddedItem {
-                        id: _email_obj_id,
-                        index: _,
-                    } in added
-                    {
-                        // [ref:TODO] do something with added items
-                    }
-                }
-                Ok(_) => {}
-                Err(err) => {
-                    log::error!(
-                        "Could not deserialize EmailQueryChangesResponse from server response:
+                    Ok(_) => {}
+                    Err(err) => {
+                        log::error!(
+                            "Could not deserialize EmailQueryChangesResponse from server response:
 - mailbox_hash: {mailbox_hash}
 - error: {err}
 - debug details:
   Json request was: {:?}
   Json reply was: {}",
-                        serde_json::to_string(&req),
-                        res_text
-                    );
-                }
-            }
-            let GetResponse::<EmailObject> { list, .. } =
-                GetResponse::<EmailObject>::try_from(v.method_responses.remove(0))?;
-            {
-                let mut mailboxes_lck = self.store.mailboxes.write().unwrap();
-                for envobj in list {
-                    if let Some(env_hash) = reverse_id_store_lck.get(&envobj.id) {
-                        let new_flags = protocol::keywords_to_flags(
-                            envobj.keywords().keys().cloned().collect(),
+                            serde_json::to_string(&req),
+                            res_text
                         );
-                        mailboxes_lck.entry(mailbox_hash).and_modify(|mbox| {
-                            if new_flags.0.contains(Flag::SEEN) {
-                                mbox.unread_emails.lock().unwrap().remove(*env_hash);
-                            } else {
-                                mbox.unread_emails.lock().unwrap().insert_new(*env_hash);
-                            }
-                        });
-                        self.add_refresh_event(RefreshEvent {
-                            account_hash: self.store.account_hash,
-                            mailbox_hash,
-                            kind: RefreshEventKind::NewFlags(*env_hash, new_flags),
-                        });
+                    }
+                }
+                let GetResponse::<EmailObject> { list, .. } =
+                    GetResponse::<EmailObject>::try_from(v.method_responses.remove(0))?;
+                {
+                    let mut mailboxes_lck = self.store.mailboxes.write().unwrap();
+                    for envobj in list {
+                        if let Some(env_hash) = reverse_id_store_lck.get(&envobj.id) {
+                            let new_flags = protocol::keywords_to_flags(
+                                envobj.keywords().keys().cloned().collect(),
+                            );
+                            mailboxes_lck.entry(mailbox_hash).and_modify(|mbox| {
+                                if new_flags.0.contains(Flag::SEEN) {
+                                    mbox.unread_emails.lock().unwrap().remove(*env_hash);
+                                } else {
+                                    mbox.unread_emails.lock().unwrap().insert_new(*env_hash);
+                                }
+                            });
+                            events.push(RefreshEvent {
+                                account_hash: self.store.account_hash,
+                                mailbox_hash,
+                                kind: RefreshEventKind::NewFlags(*env_hash, new_flags),
+                            });
+                        }
                     }
                 }
             }
@@ -688,7 +700,7 @@ impl JmapConnection {
             }
         }
 
-        Ok(())
+        Ok(events.try_into().ok())
     }
 
     pub async fn send_request(&self, request: String) -> Result<String> {
