@@ -45,22 +45,18 @@ use crate::{
 /// `/get`
 ///
 /// Objects of type `Foo` are fetched via a call to `Foo/get`.
-#[derive(Clone, Debug, Deserialize)]
-#[serde(rename_all = "camelCase")]
+#[derive(Clone, Debug)]
 pub struct Get<OBJ>
 where
     OBJ: Object + std::fmt::Debug + Serialize,
 {
     /// `account_id`: `Id` The id of the account to use.
     pub account_id: Id<Account>,
-    #[serde(skip_serializing_if = "Option::is_none")]
-    #[serde(flatten)]
     /// `ids`: `Id[]|null` The ids of the `Foo` objects to return. If `null`,
     /// then *all* records of the data type are returned, if this is
     /// supported for that data type and the number of records does not
     /// exceed the `maxObjectsInGet` limit.
     pub ids: Option<Argument<Vec<Id<OBJ>>>>,
-    #[serde(skip_serializing_if = "Option::is_none")]
     /// `properties`: `String[]|null` If supplied, only the properties listed in
     /// the array are returned for each `Foo` object. If `null`, all
     /// properties of the object are returned. The `id` property of the
@@ -68,7 +64,6 @@ where
     /// invalid property is requested, the call **MUST** be rejected with an
     /// `invalidArguments` error.
     pub properties: Option<Vec<String>>,
-    #[serde(skip)]
     _ph: PhantomData<fn() -> OBJ>,
 }
 
@@ -118,10 +113,7 @@ impl<OBJ: Object + Serialize + std::fmt::Debug> Serialize for Get<OBJ> {
     where
         S: Serializer,
     {
-        let mut fields_no = 0;
-        if !self.account_id.is_empty() {
-            fields_no += 1;
-        }
+        let mut fields_no = 1;
         if self.ids.is_some() {
             fields_no += 1;
         }
@@ -130,9 +122,7 @@ impl<OBJ: Object + Serialize + std::fmt::Debug> Serialize for Get<OBJ> {
         }
 
         let mut state = serializer.serialize_struct("Get", fields_no)?;
-        if !self.account_id.is_empty() {
-            state.serialize_field("accountId", &self.account_id)?;
-        }
+        state.serialize_field("accountId", &self.account_id)?;
         match self.ids.as_ref() {
             None => {}
             Some(Argument::Value(ref v)) => state.serialize_field("ids", v)?,
@@ -161,14 +151,76 @@ impl<OBJ: Object + Serialize + std::fmt::Debug> Serialize for Get<OBJ> {
             }
         }
 
-        if self.properties.is_some() {
-            state.serialize_field("properties", self.properties.as_ref().unwrap())?;
+        if let Some(ref properties) = self.properties {
+            state.serialize_field("properties", properties)?;
         }
 
         state.end()
     }
 }
 
+impl<'de, OBJ: Object + ::serde::de::Deserialize<'de> + Serialize + std::fmt::Debug>
+    ::serde::de::Deserialize<'de> for Get<OBJ>
+{
+    fn deserialize<D>(deserializer: D) -> std::result::Result<Self, D::Error>
+    where
+        D: ::serde::de::Deserializer<'de>,
+    {
+        #[derive(Deserialize)]
+        #[serde(rename_all = "camelCase")]
+        struct ResultReference {
+            result_of: String,
+            name: String,
+            path: String,
+        }
+
+        #[derive(Deserialize)]
+        #[serde(rename_all = "camelCase")]
+        struct Blanket<OBJ> {
+            account_id: Id<Account>,
+            #[serde(default = "default_none", rename = "#ids")]
+            _ids: Option<ResultReference>,
+            #[serde(default = "default_none")]
+            ids: Option<Vec<Id<OBJ>>>,
+            #[serde(default)]
+            properties: Option<Vec<String>>,
+        }
+
+        let Blanket {
+            account_id,
+            _ids,
+            ids,
+            properties,
+        } = Blanket::deserialize(deserializer)?;
+        match (_ids, ids) {
+            (Some(_), Some(_)) => Err(serde::de::Error::duplicate_field("ids")),
+            (Some(result_ref), None) => {
+                let ResultReference {
+                    result_of,
+                    name,
+                    path,
+                } = result_ref;
+                Ok(Self {
+                    account_id,
+                    ids: Some(Argument::ResultReference {
+                        result_of,
+                        name,
+                        path,
+                    }),
+                    properties,
+                    _ph: PhantomData,
+                })
+            }
+            (None, ids) => Ok(Self {
+                account_id,
+                ids: ids.map(Argument::Value),
+                properties,
+                _ph: PhantomData,
+            }),
+        }
+    }
+}
+
 #[derive(Clone, Debug, Deserialize, Serialize)]
 #[serde(rename_all = "camelCase")]
 pub struct MethodResponse<'a> {
@@ -215,7 +267,7 @@ enum JmapError {
     InvalidResultReference,
 }
 
-#[derive(Clone, Debug, Serialize)]
+#[derive(Clone, Debug, Deserialize, Serialize)]
 #[serde(rename_all = "camelCase")]
 pub struct Query<F: filters::FilterTrait<OBJ>, OBJ>
 where
@@ -233,10 +285,10 @@ where
     pub anchor_offset: u64,
     #[serde(skip_serializing_if = "Option::is_none")]
     pub limit: Option<u64>,
-    #[serde(default = "false_val")]
+    #[serde(default = "crate::false_val")]
     pub calculate_total: bool,
     #[serde(skip)]
-    _ph: PhantomData<fn() -> OBJ>,
+    pub _ph: PhantomData<fn() -> OBJ>,
 }
 
 impl<F: filters::FilterTrait<OBJ>, OBJ> Query<F, OBJ>
@@ -289,11 +341,11 @@ pub struct QueryResponse<OBJ: Object> {
     pub position: u64,
     pub ids: Vec<Id<OBJ>>,
     #[serde(default)]
-    pub total: u64,
+    pub total: Option<u64>,
     #[serde(default)]
-    pub limit: u64,
+    pub limit: Option<u64>,
     #[serde(skip)]
-    _ph: PhantomData<fn() -> OBJ>,
+    pub _ph: PhantomData<fn() -> OBJ>,
 }
 
 impl<OBJ: Object + DeserializeOwned> std::convert::TryFrom<&RawValue> for QueryResponse<OBJ> {
@@ -582,9 +634,15 @@ impl<OBJ: Object + Serialize + std::fmt::Debug> Serialize for Set<OBJ> {
 
         let mut state = serializer.serialize_struct("Set", fields_no)?;
         state.serialize_field("accountId", &self.account_id)?;
-        state.serialize_field("ifInState", &self.if_in_state)?;
-        state.serialize_field("update", &self.update)?;
-        state.serialize_field("destroy", &self.destroy)?;
+        if let Some(ref if_in_state) = self.if_in_state {
+            state.serialize_field("ifInState", if_in_state)?;
+        }
+        if let Some(ref update) = self.update {
+            state.serialize_field("update", update)?;
+        }
+        if let Some(ref destroy) = self.destroy {
+            state.serialize_field("destroy", destroy)?;
+        }
         if let Some(ref m) = self.create {
             let map = m
                 .into_iter()
@@ -600,8 +658,6 @@ impl<OBJ: Object + Serialize + std::fmt::Debug> Serialize for Set<OBJ> {
                 })
                 .collect::<IndexMap<_, Value>>();
             state.serialize_field("create", &map)?;
-        } else {
-            state.serialize_field("create", &self.create)?;
         }
 
         state.end()
@@ -616,6 +672,7 @@ pub struct SetResponse<OBJ: Object> {
     /// `oldState`: `String|null` The state string that would have been returned
     /// by `Foo/get` before making the requested changes, or null if the server
     /// doesn't know what the previous state string was.
+    #[serde(default = "default_none", skip_serializing_if = "Option::is_none")]
     pub old_state: Option<State<OBJ>>,
     /// `newState`: `String` The state string that will now be returned by
     /// `Foo/get`.
@@ -627,6 +684,7 @@ pub struct SetResponse<OBJ: Object> {
     /// and thus set to a default by the server.
     ///
     /// This argument is null if no `Foo` objects were successfully created.
+    #[serde(default = "default_none", skip_serializing_if = "Option::is_none")]
     pub created: Option<IndexMap<Id<OBJ>, OBJ>>,
     /// `updated`: `Id[Foo|null]|null` The keys in this map are the ids of all
     /// `Foo`s that were successfully updated.
@@ -637,22 +695,31 @@ pub struct SetResponse<OBJ: Object> {
     /// any changes to server-set or computed properties.
     ///
     /// This argument is null if no `Foo` objects were successfully updated.
+    #[serde(default = "default_none", skip_serializing_if = "Option::is_none")]
     pub updated: Option<IndexMap<Id<OBJ>, Option<OBJ>>>,
     /// `destroyed`: `Id[]|null` A list of `Foo` ids for records that were
     /// successfully destroyed, or `null` if none.
+    #[serde(default = "default_none", skip_serializing_if = "Option::is_none")]
     pub destroyed: Option<Vec<Id<OBJ>>>,
     /// `notCreated`: `Id[SetError]|null` A map of the creation id to a
     /// [`SetError`] object for each record that failed to be created, or `null`
     /// if all successful.
-    pub not_created: Option<Vec<SetError>>,
+    #[serde(default = "default_none", skip_serializing_if = "Option::is_none")]
+    pub not_created: Option<IndexMap<Id<OBJ>, SetError>>,
     /// `notUpdated`: `Id[SetError]|null` A map of the `Foo` id to a
     /// [`SetError`] object for each record that failed to be updated, or `null`
     /// if all successful.
-    pub not_updated: Option<Vec<SetError>>,
+    #[serde(default = "default_none", skip_serializing_if = "Option::is_none")]
+    pub not_updated: Option<IndexMap<Id<OBJ>, SetError>>,
     /// `notDestroyed`: `Id[SetError]|null` A map of the `Foo` id to a
     /// [`SetError`] object for each record that failed to be destroyed, or
     /// `null` if all successful.//
-    pub not_destroyed: Option<Vec<SetError>>,
+    #[serde(default = "default_none", skip_serializing_if = "Option::is_none")]
+    pub not_destroyed: Option<IndexMap<Id<OBJ>, SetError>>,
+}
+
+const fn default_none<T>() -> Option<T> {
+    None
 }
 
 impl<OBJ: Object + DeserializeOwned> std::convert::TryFrom<&RawValue> for SetResponse<OBJ> {
@@ -665,9 +732,7 @@ impl<OBJ: Object + DeserializeOwned> std::convert::TryFrom<&RawValue> for SetRes
     }
 }
 
-#[derive(Clone, Debug, Deserialize, Serialize)]
-#[serde(rename_all = "camelCase")]
-#[serde(tag = "type", content = "description")]
+#[derive(Clone, Debug)]
 pub enum SetError {
     /// (create; update; destroy). The create/update/destroy would violate an
     /// ACL or other permissions policy.
@@ -709,6 +774,123 @@ pub enum SetError {
     StateMismatch(Option<String>),
 }
 
+impl Serialize for SetError {
+    fn serialize<S>(&self, serializer: S) -> std::result::Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let (r#type, description) = match self {
+            Self::Forbidden(ref d) => ("forbidden", d),
+            Self::OverQuota(ref d) => ("overQuota", d),
+            Self::TooLarge(ref d) => ("tooLarge", d),
+            Self::RateLimit(ref d) => ("rateLimit", d),
+            Self::NotFound(ref d) => ("notFound", d),
+            Self::InvalidPatch(ref d) => ("invalidPatch", d),
+            Self::WillDestroy(ref d) => ("willDestroy", d),
+            Self::InvalidProperties {
+                ref description,
+                ref properties,
+            } => {
+                if let Some(description) = description {
+                    let mut set_error = serializer.serialize_struct("SetError", 3)?;
+                    set_error.serialize_field("type", "invalidProperties")?;
+                    set_error.serialize_field("description", description)?;
+                    set_error.serialize_field("properties", properties)?;
+                    return set_error.end();
+                } else {
+                    let mut set_error = serializer.serialize_struct("SetError", 2)?;
+                    set_error.serialize_field("type", "invalidProperties")?;
+                    set_error.serialize_field("properties", properties)?;
+                    return set_error.end();
+                }
+            }
+            Self::Singleton(ref d) => ("singleton", d),
+            Self::RequestTooLarge(ref d) => ("requestTooLarge", d),
+            Self::StateMismatch(ref d) => ("stateMismatch", d),
+        };
+        if let Some(description) = description {
+            let mut set_error = serializer.serialize_struct("SetError", 2)?;
+            set_error.serialize_field("type", r#type)?;
+            set_error.serialize_field("description", description)?;
+            set_error.end()
+        } else {
+            let mut set_error = serializer.serialize_struct("SetError", 1)?;
+            set_error.serialize_field("type", r#type)?;
+            set_error.end()
+        }
+    }
+}
+
+impl<'de> ::serde::de::Deserialize<'de> for SetError {
+    fn deserialize<D>(deserializer: D) -> std::result::Result<Self, D::Error>
+    where
+        D: ::serde::de::Deserializer<'de>,
+    {
+        use serde::de::Error;
+
+        #[derive(Deserialize)]
+        struct Blanket {
+            r#type: String,
+            #[serde(default)]
+            description: Option<String>,
+            #[serde(default)]
+            properties: Option<Vec<String>>,
+        }
+
+        let Blanket {
+            r#type,
+            description,
+            properties,
+        } = Blanket::deserialize(deserializer)?;
+        if r#type != "invalidProperties" && properties.is_some() {
+            return Err(serde::de::Error::unknown_field(
+                "properties",
+                &["type", "description"],
+            ));
+        }
+        let set_error = match r#type.as_str() {
+            "forbidden" => Self::Forbidden(description),
+            "overQuota" => Self::OverQuota(description),
+            "tooLarge" => Self::TooLarge(description),
+            "rateLimit" => Self::RateLimit(description),
+            "notFound" => Self::NotFound(description),
+            "invalidPatch" => Self::InvalidPatch(description),
+            "willDestroy" => Self::WillDestroy(description),
+            "invalidProperties" => {
+                let Some(properties) = properties else {
+                    return Err(Error::missing_field("properties"));
+                };
+                Self::InvalidProperties {
+                    description,
+                    properties,
+                }
+            }
+            "singleton" => Self::Singleton(description),
+            "requestTooLarge" => Self::RequestTooLarge(description),
+            "stateMismatch" => Self::StateMismatch(description),
+            other => {
+                return Err(Error::unknown_variant(
+                    other,
+                    &[
+                        "forbidden",
+                        "overQuota",
+                        "tooLarge",
+                        "rateLimit",
+                        "notFound",
+                        "invalidPatch",
+                        "willDestroy",
+                        "invalidProperties",
+                        "singleton",
+                        "requestTooLarge",
+                        "stateMismatch",
+                    ],
+                ));
+            }
+        };
+        Ok(set_error)
+    }
+}
+
 impl std::fmt::Display for SetError {
     fn fmt(&self, fmt: &mut std::fmt::Formatter) -> std::fmt::Result {
         use SetError::*;
@@ -798,6 +980,15 @@ impl<'de> ::serde::de::Deserialize<'de> for RequestUrlTemplate {
     }
 }
 
+#[allow(clippy::literal_string_with_formatting_args)]
+const ACCOUNT_ID_URL_FORMAT_ARG: &str = "{accountId}";
+#[allow(clippy::literal_string_with_formatting_args)]
+const BLOB_ID_URL_FORMAT_ARG: &str = "{blobId}";
+#[allow(clippy::literal_string_with_formatting_args)]
+const NAME_URL_FORMAT_ARG: &str = "{name}";
+#[allow(clippy::literal_string_with_formatting_args)]
+const TYPE_URL_FORMAT_ARG: &str = "{type}";
+
 pub fn download_request_format(
     download_url: &RequestUrlTemplate,
     account_id: &Id<Account>,
@@ -815,18 +1006,18 @@ pub fn download_request_format(
     while let Some(pos) = download_url.text.as_bytes()[prev_pos..].find(b"{") {
         ret.push_str(&download_url.text[prev_pos..prev_pos + pos]);
         prev_pos += pos;
-        if download_url.text[prev_pos..].starts_with("{accountId}") {
+        if download_url.text[prev_pos..].starts_with(ACCOUNT_ID_URL_FORMAT_ARG) {
             ret.push_str(account_id.as_str());
-            prev_pos += "{accountId}".len();
-        } else if download_url.text[prev_pos..].starts_with("{blobId}") {
+            prev_pos += ACCOUNT_ID_URL_FORMAT_ARG.len();
+        } else if download_url.text[prev_pos..].starts_with(BLOB_ID_URL_FORMAT_ARG) {
             ret.push_str(blob_id.as_str());
-            prev_pos += "{blobId}".len();
-        } else if download_url.text[prev_pos..].starts_with("{name}") {
+            prev_pos += BLOB_ID_URL_FORMAT_ARG.len();
+        } else if download_url.text[prev_pos..].starts_with(NAME_URL_FORMAT_ARG) {
             ret.push_str(name.as_deref().unwrap_or(""));
-            prev_pos += "{name}".len();
-        } else if download_url.text[prev_pos..].starts_with("{type}") {
+            prev_pos += NAME_URL_FORMAT_ARG.len();
+        } else if download_url.text[prev_pos..].starts_with(TYPE_URL_FORMAT_ARG) {
             ret.push_str("application/octet-stream");
-            prev_pos += "{name}".len();
+            prev_pos += TYPE_URL_FORMAT_ARG.len();
         } else {
             log::error!(
                 "BUG: unknown parameter in download_url: {}",
@@ -877,9 +1068,9 @@ pub fn upload_request_format(
     while let Some(pos) = upload_url.text.as_bytes()[prev_pos..].find(b"{") {
         ret.push_str(&upload_url.text[prev_pos..prev_pos + pos]);
         prev_pos += pos;
-        if upload_url.text[prev_pos..].starts_with("{accountId}") {
+        if upload_url.text[prev_pos..].starts_with(ACCOUNT_ID_URL_FORMAT_ARG) {
             ret.push_str(account_id.as_str());
-            prev_pos += "{accountId}".len();
+            prev_pos += ACCOUNT_ID_URL_FORMAT_ARG.len();
             break;
         } else {
             ret.push('{');
