@@ -296,6 +296,7 @@ pub struct Store {
     pub online_status: OnlineStatus,
     pub is_subscribed: IsSubscribedFn,
     pub core_capabilities: Arc<Mutex<IndexMap<String, CapabilitiesObject>>>,
+    pub metadata: Arc<Mutex<IndexMap<String, serde_json::Value>>>,
     pub event_consumer: BackendEventConsumer,
 }
 
@@ -395,18 +396,51 @@ impl MailBackend for JmapType {
             supports_search: true,
             extensions: None,
             supports_tags: true,
-            supports_submission: true,
+            supports_submission: false,
             extra_submission_headers: &[],
             metadata: None,
         };
-        let supports_submission: bool = self
-            .store
-            .core_capabilities
-            .lock()
-            .map(|c| c.contains_key(JmapSubmissionCapability::uri()))
-            .unwrap_or(false);
+        let mut supports_submission = false;
+        let mut extensions = None;
+        if let Ok(core_capabilities) = self.store.core_capabilities.lock() {
+            let mut caps = vec![];
+            supports_submission = core_capabilities.contains_key(JmapSubmissionCapability::uri());
+            for k in core_capabilities.keys() {
+                if [
+                    JmapCoreCapability::uri(),
+                    JmapMailCapability::uri(),
+                    JmapSubmissionCapability::uri(),
+                ]
+                .contains(&k.as_str())
+                {
+                    caps.push((
+                        k.to_string(),
+                        MailBackendExtensionStatus::Supported { comment: None },
+                    ));
+                } else {
+                    caps.push((
+                        k.to_string(),
+                        MailBackendExtensionStatus::Unsupported { comment: None },
+                    ));
+                }
+            }
+            if !caps.is_empty() {
+                extensions = Some(caps);
+            }
+        };
+        let metadata = {
+            let lck = self.store.metadata.lock().unwrap();
+            if lck.is_empty() {
+                None
+            } else {
+                Some(serde_json::json! {lck.clone()})
+            }
+        };
+
         MailBackendCapabilities {
-            supports_submission: CAPABILITIES.supports_submission || supports_submission,
+            supports_submission,
+            metadata,
+            extensions,
             ..CAPABILITIES
         }
     }
@@ -452,15 +486,17 @@ impl MailBackend for JmapType {
         Ok(Box::pin(async move {
             let mut conn = connection.lock().await;
             conn.connect().await?;
-            conn.email_changes(mailbox_hash).await?;
+            if let Some(ev) = conn.email_changes(mailbox_hash).await? {
+                conn.add_backend_event(ev);
+            }
             Ok(())
         }))
     }
 
-    fn watch(&self) -> ResultFuture<()> {
+    fn watch(&self) -> ResultStream<BackendEvent> {
         let connection = self.connection.clone();
         let store = self.store.clone();
-        Ok(Box::pin(async move {
+        Ok(Box::pin(try_fn_stream(|emitter| async move {
             {
                 let mut conn = connection.lock().await;
                 conn.connect().await?;
@@ -478,12 +514,14 @@ impl MailBackend for JmapType {
                     };
                     let conn = connection.lock().await;
                     for mailbox_hash in mailbox_hashes {
-                        conn.email_changes(mailbox_hash).await?;
+                        if let Some(ev) = conn.email_changes(mailbox_hash).await? {
+                            emitter.emit(ev).await;
+                        }
                     }
                 }
                 sleep(Duration::from_secs(60)).await;
             }
-        }))
+        })))
     }
 
     fn mailboxes(&self) -> ResultFuture<HashMap<MailboxHash, Mailbox>> {
@@ -509,12 +547,9 @@ impl MailBackend for JmapType {
         }))
     }
 
-    fn operation(&self, hash: EnvelopeHash) -> Result<Box<dyn BackendOp>> {
-        Ok(Box::new(JmapOp::new(
-            hash,
-            self.connection.clone(),
-            self.store.clone(),
-        )))
+    fn envelope_bytes_by_hash(&self, hash: EnvelopeHash) -> ResultFuture<Vec<u8>> {
+        let op = JmapOp::new(hash, self.connection.clone(), self.store.clone());
+        Ok(Box::pin(async move { op.as_bytes().await }))
     }
 
     fn save(
@@ -634,11 +669,8 @@ impl MailBackend for JmapType {
                 .id
                 .clone();
 
-            let mut f = Filter::Condition(
-                email::EmailFilterCondition::new()
-                    .in_mailbox(Some(mailbox_id))
-                    .into(),
-            );
+            let mut f =
+                Filter::Condition(email::EmailFilterCondition::new().in_mailbox(Some(mailbox_id)));
             f &= Filter::<email::EmailFilterCondition, email::EmailObject>::from(q);
             f
         } else {
@@ -716,6 +748,7 @@ impl MailBackend for JmapType {
 
             let mut req = Request::new(conn.request_no.clone());
             let _prev_seq = req.add_call(&mailbox_set_call).await;
+            // [ref:FIXME]: inspect Set response for errors in `notCreated` field.
             let new_mailboxes = protocol::get_mailboxes(&mut conn, Some(req)).await?;
 
             let new_mailbox: Mailbox = new_mailboxes
@@ -760,6 +793,7 @@ impl MailBackend for JmapType {
 
             let mut req = Request::new(conn.request_no.clone());
             let _prev_seq = req.add_call(&mailbox_set_call).await;
+            // [ref:FIXME]: inspect Set response for errors in `notCreated` field.
             let new_mailboxes = protocol::get_mailboxes(&mut conn, Some(req)).await?;
             *store.mailboxes.write().unwrap() = new_mailboxes;
 
@@ -834,18 +868,21 @@ impl MailBackend for JmapType {
                 if !ids.is_empty() {
                     return Err(Error::new(format!(
                         "Could not delete mailbox: {}",
-                        ids.iter()
+                        ids.values()
                             .map(|err| err.to_string())
                             .collect::<Vec<String>>()
                             .join(",")
                     )));
                 }
             }
-            conn.add_refresh_event(RefreshEvent {
-                account_hash: store.account_hash,
-                mailbox_hash,
-                kind: RefreshEventKind::MailboxDelete(mailbox_hash),
-            });
+            conn.add_backend_event(
+                RefreshEvent {
+                    account_hash: store.account_hash,
+                    mailbox_hash,
+                    kind: RefreshEventKind::MailboxDelete(mailbox_hash),
+                }
+                .into(),
+            );
             let new_mailboxes = protocol::get_mailboxes(&mut conn, None).await?;
             *store.mailboxes.write().unwrap() = new_mailboxes;
 
@@ -898,12 +935,26 @@ impl MailBackend for JmapType {
             let res_text = conn.send_request(serde_json::to_string(&req)?).await?;
             let v: MethodResponse = deserialize_from_str(&res_text)?;
             conn.store.online_status.update_timestamp(None).await;
-            let SetResponse::<mailbox::MailboxObject> { new_state, .. } =
-                SetResponse::<mailbox::MailboxObject>::try_from(
-                    *v.method_responses.last().unwrap(),
-                )?;
+            let SetResponse::<mailbox::MailboxObject> {
+                new_state,
+                not_updated,
+                ..
+            } = SetResponse::<mailbox::MailboxObject>::try_from(
+                *v.method_responses.last().unwrap(),
+            )?;
             *conn.store.mailbox_state.lock().await = Some(new_state);
             conn.last_method_response = Some(res_text);
+            if let Some(ids) = not_updated {
+                if !ids.is_empty() {
+                    return Err(Error::new(format!(
+                        "Could not update mailbox subscription: {}",
+                        ids.values()
+                            .map(|err| err.to_string())
+                            .collect::<Vec<String>>()
+                            .join(",")
+                    )));
+                }
+            }
 
             Ok(())
         }))
@@ -966,8 +1017,6 @@ impl MailBackend for JmapType {
             {
                 for env_hash in env_hashes.iter() {
                     if let Some(id) = store.id_store.lock().await.get(&env_hash) {
-                        // ids.push(id.clone());
-                        // id_map.insert(id.clone(), env_hash);
                         update_map.insert(
                             Argument::from(id.clone()),
                             serde_json::json!(update_keywords.clone()),
@@ -975,6 +1024,7 @@ impl MailBackend for JmapType {
                     }
                 }
             }
+            let batch_len = update_map.len();
             let conn = connection.lock().await;
             let mail_account_id = conn.session_guard().await?.mail_account_id();
             let state = conn.store.email_state.lock().await.clone();
@@ -1011,8 +1061,9 @@ impl MailBackend for JmapType {
             if let Some(ids) = not_updated {
                 if !ids.is_empty() {
                     return Err(Error::new(format!(
-                        "Could not update ids: {}",
-                        ids.into_iter()
+                        "Could not copy {}emails: {}",
+                        if ids.len() == batch_len { "" } else { "some " },
+                        ids.values()
                             .map(|err| err.to_string())
                             .collect::<Vec<String>>()
                             .join(",")
@@ -1034,6 +1085,8 @@ impl MailBackend for JmapType {
         Ok(Box::pin(async move {
             let mut update_map: IndexMap<Argument<Id<email::EmailObject>>, Value> =
                 IndexMap::default();
+            let mut updated_map: IndexMap<Id<email::EmailObject>, EnvelopeHash> =
+                IndexMap::default();
             let mut ids: Vec<Id<email::EmailObject>> =
                 Vec::with_capacity(env_hashes.rest.len() + 1);
             let mut id_map: IndexMap<Id<email::EmailObject>, EnvelopeHash> = IndexMap::default();
@@ -1083,17 +1136,19 @@ impl MailBackend for JmapType {
                 }
             }
             {
-                for hash in env_hashes.iter() {
-                    if let Some(id) = store.id_store.lock().await.get(&hash) {
+                for env_hash in env_hashes.iter() {
+                    if let Some(id) = store.id_store.lock().await.get(&env_hash) {
                         ids.push(id.clone());
-                        id_map.insert(id.clone(), hash);
+                        id_map.insert(id.clone(), env_hash);
                         update_map.insert(
                             Argument::from(id.clone()),
                             serde_json::json!(update_keywords.clone()),
                         );
+                        updated_map.insert(id.clone(), env_hash);
                     }
                 }
             }
+            let batch_len = update_map.len();
             let conn = connection.lock().await;
             let mail_account_id = conn.session_guard().await?.mail_account_id();
             let state = conn.store.email_state.lock().await.clone();
@@ -1134,17 +1189,14 @@ impl MailBackend for JmapType {
                 new_state,
                 ..
             } = SetResponse::<email::EmailObject>::try_from(v.method_responses.remove(0))?;
-            *conn.store.email_state.lock().await = Some(new_state);
-            if let Some(ids) = not_updated {
-                return Err(Error::new(
-                    ids.into_iter()
-                        .map(|err| err.to_string())
-                        .collect::<Vec<String>>()
-                        .join(","),
-                ));
+            if let Some(ref ids) = not_updated {
+                for id in ids.keys() {
+                    updated_map.swap_remove(id);
+                }
             }
 
-            {
+            *conn.store.email_state.lock().await = Some(new_state);
+            if !updated_map.is_empty() {
                 let mut tag_index_lck = store.collection.tag_index.write().unwrap();
                 for op in flags.iter() {
                     if let FlagOp::SetTag(t) = op {
@@ -1153,20 +1205,43 @@ impl MailBackend for JmapType {
                 }
                 drop(tag_index_lck);
             }
-            let e = GetResponse::<email::EmailObject>::try_from(v.method_responses.pop().unwrap())?;
-            let GetResponse::<email::EmailObject> { list, .. } = e;
-            debug!(&list);
-            for envobj in list {
-                let env_hash = id_map[&envobj.id];
-                conn.add_refresh_event(RefreshEvent {
-                    account_hash: store.account_hash,
-                    mailbox_hash,
-                    kind: RefreshEventKind::NewFlags(
-                        env_hash,
-                        protocol::keywords_to_flags(envobj.keywords().keys().cloned().collect()),
-                    ),
-                });
+            let GetResponse::<email::EmailObject> { list, .. } =
+                GetResponse::try_from(v.method_responses.pop().unwrap())?;
+            if !list.is_empty() {
+                let list = list
+                    .into_iter()
+                    .filter(|envobj| updated_map.contains_key(&envobj.id))
+                    .map(|envobj| {
+                        let env_hash = id_map[&envobj.id];
+                        RefreshEvent {
+                            account_hash: store.account_hash,
+                            mailbox_hash,
+                            kind: RefreshEventKind::NewFlags(
+                                env_hash,
+                                protocol::keywords_to_flags(
+                                    envobj.keywords().keys().cloned().collect(),
+                                ),
+                            ),
+                        }
+                    })
+                    .collect::<Vec<_>>();
+                if let Ok(ev) = list.try_into() {
+                    conn.add_backend_event(ev);
+                }
             }
+            if let Some(ids) = not_updated {
+                if !ids.is_empty() {
+                    return Err(Error::new(format!(
+                        "Could not update {}email flags: {}",
+                        if ids.len() == batch_len { "" } else { "some " },
+                        ids.values()
+                            .map(|err| err.to_string())
+                            .collect::<Vec<String>>()
+                            .join(",")
+                    )));
+                }
+            }
+
             Ok(())
         }))
     }
@@ -1190,9 +1265,11 @@ impl MailBackend for JmapType {
         Ok(Box::pin(async move {
             let mut destroy_vec: Vec<Argument<Id<email::EmailObject>>> =
                 Vec::with_capacity(env_hashes.len());
+            let mut destroyed_map: IndexMap<Id<email::EmailObject>, EnvelopeHash> = IndexMap::new();
             for env_hash in env_hashes.iter() {
                 if let Some(id) = store.id_store.lock().await.get(&env_hash) {
                     destroy_vec.push(Argument::Value(id.clone()));
+                    destroyed_map.insert(id.clone(), env_hash);
                 }
             }
             let batch_len = destroy_vec.len();
@@ -1229,22 +1306,29 @@ impl MailBackend for JmapType {
                 ..
             } = SetResponse::<email::EmailObject>::try_from(v.method_responses.remove(0))?;
             *conn.store.email_state.lock().await = Some(new_state);
-            conn.add_backend_event(BackendEvent::RefreshBatch(
-                env_hashes
-                    .iter()
-                    .map(|h| RefreshEvent {
-                        account_hash: store.account_hash,
-                        mailbox_hash,
-                        kind: RefreshEventKind::Remove(h),
-                    })
-                    .collect(),
-            ));
+            if let Some(ref ids) = not_destroyed {
+                for id in ids.keys() {
+                    destroyed_map.swap_remove(id);
+                }
+            }
+            if !destroyed_map.is_empty() {
+                conn.add_backend_event(BackendEvent::RefreshBatch(
+                    destroyed_map
+                        .values()
+                        .map(|h| RefreshEvent {
+                            account_hash: store.account_hash,
+                            mailbox_hash,
+                            kind: RefreshEventKind::Remove(*h),
+                        })
+                        .collect(),
+                ));
+            }
             if let Some(ids) = not_destroyed {
                 if !ids.is_empty() {
                     return Err(Error::new(format!(
-                        "Could not destroy {}ids: {}",
+                        "Could not delete {}emails: {}",
                         if ids.len() == batch_len { "" } else { "some " },
-                        ids.iter()
+                        ids.values()
                             .map(|err| err.to_string())
                             .collect::<Vec<String>>()
                             .join(",")
@@ -1449,6 +1533,7 @@ impl JmapType {
             is_subscribed,
             collection: Collection::default(),
             core_capabilities: Arc::new(Mutex::new(IndexMap::default())),
+            metadata: Arc::new(Mutex::new(IndexMap::default())),
             byte_cache: Default::default(),
             id_store: Default::default(),
             reverse_id_store: Default::default(),
