@@ -741,7 +741,7 @@ impl SmtpConnection {
         // given above.
 
         if self.server_conf.extensions.binarymime {
-            let mail_length = format!("{}", mail.as_bytes().len());
+            let mail_length = format!("{}", mail.len());
             self.send_command(&[b"BDAT", mail_length.as_bytes(), b"LAST"])
                 .await?;
             self.stream.write_all(mail.as_bytes()).await?;
@@ -910,6 +910,7 @@ pub enum ReplyCode {
 }
 
 impl ReplyCode {
+    /// Return a description of the reply code in English.
     pub const fn as_str(&self) -> &'static str {
         use ReplyCode::*;
         match self {
@@ -951,6 +952,43 @@ impl ReplyCode {
         }
     }
 
+    /// Return the reply code as an integer.
+    pub const fn value(&self) -> u16 {
+        use ReplyCode::*;
+        match self {
+            _211 => 211,
+            _214 => 214,
+            _220 => 220,
+            _221 => 221,
+            _250 => 250,
+            _235 => 235,
+            _251 => 251,
+            _252 => 252,
+            _334 => 334,
+            _353 => 353,
+            _354 => 354,
+            _421 => 421,
+            _450 => 450,
+            _451 => 451,
+            _452 => 452,
+            _455 => 455,
+            _500 => 500,
+            _501 => 501,
+            _502 => 502,
+            _503 => 503,
+            _504 => 504,
+            _535 => 535,
+            _550 => 550,
+            _551 => 551,
+            _552 => 552,
+            _553 => 553,
+            _554 => 554,
+            _555 => 555,
+            _530 => 530,
+        }
+    }
+
+    /// Returns `true` if reply code indicates that an error has occured.
     pub const fn is_err(&self) -> bool {
         use ReplyCode::*;
         matches!(
@@ -976,6 +1014,7 @@ impl ReplyCode {
     }
 }
 
+/// Parse a [`ReplyCode`] from a string slice of three digits, e.g. `"250"`.
 impl TryFrom<&'_ str> for ReplyCode {
     type Error = Error;
     fn try_from(val: &'_ str) -> Result<Self> {
